/******************************************************************************
 **																			 **
 ** 	Module:		DriverBird.c					    					 **
 ** 																		 **
 ** 																		 **
 ** 	Purpose: 	Connects to the Ascension Flock Of Birds				 **
 **					Adapted for Flock Of Birds by Maarten Gribnau			 **
 **					Would have liked to use the reference frame commands of	 **
 **					the flock for translating from flock to QD3D coordinate	 **
 **					system but it did not work out due to the flock's sensi- **
 **					tiveness to angles approaching 90 degrees				 **
 **					email: m.w.gribnau@io.tudelft.nl						 **
 **					The original notice is included below.					 **
 **					Based on code for Magellan driver.						 **
 ** 				Copyright (C) 1992-96 Apple Computer, Inc.				 **
 **					All rights reserved.									 **
 ** 																		 **
 ** 																		 **
 ** 																		 **
 ** 																		 **
 *****************************************************************************/

/* this code is supplied on the understanding that it is sample code
 * and is completly unfit for any particular purpose.  In addition, this
 * sample is unsupported.
 *
 * Please send questions to DEVSUPPORT, send bugs to NICKT
 *
 * This project builds an app which is the "driver" app.  You need to 
 * run this before any application that uses this driver.  This app will 
 * sit there and poll the input device for changes.  When it receives changes
 * in orientation it will update the controller.  The app that uses the controller
 * (a sample is supplied with this driver, the sample is called tryTracker) must
 * poll the controller to get the orientation changes.
 *
 * Please contact stephen ilberg at Logitech for more information on the Magellan product
 */


#include "DriverBird.h"

#include <Devices.h>
#include <Errors.h>
#include <Serial.h>

#include <QD3DController.h>
#include <QD3DMath.h>

#include "SerialDriverNames.h"
#include "SerialDriverArbitration.h"
#include "DriverErrors.h"

#define MAX_NUM_ADDRESSES 16
#define MAX_NUM_CONTROLLERS MAX_NUM_ADDRESSES
#define PACKET_SIZE_STAND_ALONE 15 /* 14 position/quaternion, 1 button byte */
#define PACKET_SIZE_MASTER_SLAVE 16 /* 1 address byte is added at end */
#define BUFFER_LENGTH 256

/******************************************************************************
 **																			 **
 **							Variables and Forward Declarations				 **
 **																			 **
 *****************************************************************************/


static TQ3ControllerRef		fControllerRefs[MAX_NUM_CONTROLLERS];
static short				fNumControllers = 0;

static short				fPort = -1;
static short				fSerInp = 0;
static short				fSerOut = 0;
static short				fPortSpeed = baud9600;

static DeferredTask			fDefTask;
static IOParam				fSerParm;
static Byte					fBuffer[BUFFER_LENGTH];
static long					fBufferCount;

static short				fBirdAddresses[MAX_NUM_ADDRESSES];
static short				fNumBirds = 0;
static Boolean				fStandAlone = true;
static long					fPacketSize;
static TBirdSync			fBirdSync = sync_None;
static TBirdRange			fBirdRange = range_36;
static short				fNumSkip = 4;
static short				fSkip = 0;
static TQ3Point3D			fPrevPoint[MAX_NUM_CONTROLLERS];
static TQ3Quaternion		fPrevQuaternion[MAX_NUM_CONTROLLERS];
static Boolean				fIsRunning = false;
#if OPERATION_MODE
static TBirdOperation		fOperation = mode_Poll;
#endif
static TBirdMode			fMode = mode_Relative;
static Boolean				fTrack2DCursor = false;
/* 1 inch equals fInch2QD3D Quickdraw 3D units*/
static float				fInch2Qd3dX = 1;
static float				fInch2Qd3dY = 1;
static float				fInch2Qd3dZ = 1;
/* Scales the bird coordinates to QD3D units */
static float				fScaleX = 1/36;
static float				fScaleY = 1/36;
static float				fScaleZ = 1/36;

/******************************************************************************
 **	Prototypes of private routines
 *****************************************************************************/

static void Wait(
	unsigned long milliSec);

static OSErr SerRecvFlush(
	short serPort);

static OSErr SerSendData(
	Byte *data,
	long count);

static OSErr SerRecvData(
	Byte *data,
	int	requested);

static pascal void ASyncCompletion(
	ParmBlkPtr paramBlock);

static pascal void DeferredCompletion(
	long dtParam);

static OSErr ControllerInit(
	short numControllers);

static void ControllerTerm(
	void);

static short ControllerFind(
	Byte address);

static OSErr SerialPortInit(
	short portID,
	short speed,
	short *serPortIn,
	short *serPortOut);
static OSErr SerialPortTerm(
	short serPortIn,
	short serPortOut);

#if 0
static void EncodeValue(
	float value,
	Byte *lsb,
	Byte *msb);

static void EncodeAngle(
	float angle,
	Byte *lsb,
	Byte *msb);

static void EncodeQD3DReferenceFrame1(
	Byte bytes[12]);
#endif

static void DecodeValue(
	Byte lsb,
	Byte msb,
	float *value);

static void	DecodePoint(
	Byte bytes[6],
	TQ3Point3D *point);

static void	DecodeQuaternion(
	Byte bytes[8],
	TQ3Quaternion *quaternion);

static void DecodeButtons(
	Byte byte,
	unsigned long *buttons);

static OSErr GetBirdError(
	void);

static OSErr GetBirdIsStandAlone(
	Boolean *standAlone);

static OSErr GetBirdsPresent(
	short addresses[MAX_NUM_ADDRESSES],
	short *numAddresses,
	short *numFound,
	Boolean standAlone);

static OSErr SetBirdsRunning(
	short addresses[MAX_NUM_ADDRESSES],
	short numAddresses,
	Boolean standAlone);

static void UpdateControllerRelative(
	int index,
	TQ3Point3D *prevP,
	TQ3Quaternion *prevQ,
	TQ3Point3D *currP,
	TQ3Quaternion *currQ,
	unsigned long buttons);

static void UpdateControllerAbsolute(
	int index,
	TQ3Point3D *point,
	TQ3Quaternion *quaternion,
	unsigned long buttons);


/******************************************************************************
 **	Private Routines
 *****************************************************************************/

/*===========================================================================*\
 *
 *	Routine:	Wait()
 *
 *	Comments:	Wait a given time in milli seconds
 *
\*===========================================================================*/

static void Wait(unsigned long milliSec)
{
	unsigned long end;

	end = TickCount() + (milliSec / 16.66);
	while (end > TickCount()) ;
}


/*===========================================================================*\
 *
 *	Routine:	SerRecvFlush()
 *
 *	Comments:	Clear the serial input port buffer 
 *
\*===========================================================================*/

static OSErr SerRecvFlush(
	short serPort)
{
	OSErr		err = noErr;
	long		count;
	char		buffer[BUFFER_LENGTH];
	Boolean		ready = false;
	
	while (!ready) {
		err = SerGetBuf(serPort, &count);
		if (err != noErr) break;
		if (count > 0) {
			if (count > BUFFER_LENGTH) {
				count = BUFFER_LENGTH;
			}
			err = FSRead(serPort, &count, (Ptr) buffer);
			if (err != noErr) break;
		}
		else {
			ready = true;
		}
	}
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	SerSendData()
 *
 *	Comments:	Sends the given bytes to the device.
 *
\*===========================================================================*/

static OSErr SerSendData(
	Byte *data,
	long count)
{
	return FSWrite(fSerOut, &count, (Ptr) data);	
}


/*===========================================================================*\
 *
 *	Routine:	SerRecvData()
 *
 *	Comments:	Reads the requested number of bytes from the device
 *				This routine times out after one second initially, and
 *				after one tenth second between characters.
 *
\*===========================================================================*/

static OSErr SerRecvData(
	Byte *data,
	int	requested)
{
	OSErr			err;
	long			count, total;
	unsigned long	timeout;
	Boolean			ready = false;
	
	timeout = TickCount() + 60;
	total = 0;
	
	while (!ready) {
		err = SerGetBuf(fSerInp, &count);
		if (err != noErr) break;
		
		if (count > 0) {
			if ((count + total) > requested) {
				count = requested - total;
			}
			err = FSRead(fSerInp, &count, (Ptr) data+total);
			if (err != noErr) break;
			total += count;
			if (total >= requested) {
				ready = true;
			}
			else {			
				timeout = TickCount() + 6;			
			}
		}
		else {
			if (TickCount() > timeout) {
				err = err_BirdNotResponding;
				ready = true;
			}
		}
	}
	
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	ASyncCompletion()
 *
 *	Comments:	This is the serial input async completion routine. It appends
 *				the incoming data to the buffer, and invokes a deferred task
 *				to parse it.
 *
\*===========================================================================*/

static pascal void ASyncCompletion(
	ParmBlkPtr /*paramBlock*/)
{
	if (fSerParm.ioResult == abortErr) {
		/* Bail out on KillIO */
		return;
	}
	fBufferCount += fSerParm.ioActCount;
	if (fSerParm.ioResult != noErr) {
		/* Bag the buffer on serial error */	
		fBufferCount = 0;
	}
	DTInstall(&fDefTask);
}


/*===========================================================================*\
 *
 *	Routine:	DeferredCompletion()
 *
 *	Comments:	This is the deferred task that is invoked by the serial input
 *				async completion routine. It tries to parse the data in the
 *				serial input buffer, and issues	another read for the next
 *				block of data.
 *				The flock is set in group mode, but one birds package is pro-
 *				cessed at a time. 
 *
\*===========================================================================*/

static pascal void DeferredCompletion(
	long /*dtParam*/)
{
	long count;
	Byte *buf;
	long reqCount;
	TQ3Point3D point;
	TQ3Quaternion quaternion;
	unsigned long buttons;
	short controller;

	reqCount = fPacketSize;

	buf = fBuffer;
	count = fBufferCount;
	while (count > 0) {
		if (*buf & 0x80) {
			if (count < fPacketSize) {
				/* Not enough bytes in buffer, break out of loop */
				break;
			}
			else {
				buf[0] &= 0x7F;
				DecodePoint(buf, &point);
				DecodeQuaternion(buf+6, &quaternion);
				Q3Quaternion_Normalize(&quaternion, &quaternion);
				DecodeButtons(buf[14], &buttons);
				controller = ControllerFind(buf[15]);
				count -= fPacketSize;
				buf += fPacketSize;
				if ((controller >= 0) && (controller < fNumBirds)) {
					if (fSkip >= fNumSkip) {
						if (fMode == mode_Relative) {
							UpdateControllerRelative(controller,
							                         &fPrevPoint[controller],
							                         &fPrevQuaternion[controller],
							                         &point,
							                         &quaternion,
							                         buttons);
						}
						else {
							UpdateControllerAbsolute(controller,
							                         &point,
							                         &quaternion,
							                         buttons); 
						}
					}
					else {
						fSkip++;
					}
					/* Save the current measurement */
					fPrevPoint[controller] = point;
					fPrevQuaternion[controller] = quaternion;
				}
			}
		}
		else {
			count--;
			buf++;
		}
	}
	
	/* Prepare to read the next package */
	if (count > 0) {
		/* Bytes left in the buffer */
		if (count < fPacketSize) {
			BlockMove(buf, fBuffer, count);
			fBufferCount = count;
			fSerParm.ioBuffer	= (Ptr) fBuffer;
			fSerParm.ioReqCount	= fPacketSize - count;
		}
		else {
			fSerParm.ioBuffer	= (Ptr) fBuffer;
			fSerParm.ioReqCount	= fPacketSize;
			fBufferCount = 0;
		}
	}
	else {
		fSerParm.ioBuffer	= (Ptr) fBuffer;
		fSerParm.ioReqCount	= fPacketSize;
		fBufferCount = 0;
	}
	PBReadAsync((ParmBlkPtr) &fSerParm);
}


/*===========================================================================*\
 *
 *	Routine:	ControllerInit()
 *
 *	Comments:	Creates the number of QD3D controllers requested
 *
\*===========================================================================*/

static OSErr ControllerInit(
	short numControllers)
{
	TQ3ControllerData controllerData;
	char signature[75];
	int i;
	OSErr err;
	
	err = noErr;
	if (numControllers > MAX_NUM_CONTROLLERS) {
		numControllers = MAX_NUM_CONTROLLERS;
	}
	controllerData.signature		= signature;
	controllerData.valueCount		= 0;
	controllerData.channelCount		= 0;
	controllerData.channelGetMethod	= NULL ;	/* no channels */
	controllerData.channelSetMethod	= NULL ;	/* no channels */
	for (i = 0; i < numControllers; i++) {
		sprintf(signature, "Flock Of Birds number %d, Maarten Gribnau.", i);
		fControllerRefs[i] = Q3Controller_New(&controllerData);
		if (fControllerRefs[i] == NULL) {
			err = err_Controller;
			break;
		}
	}
	fNumControllers = i;
	
	return err;	
}


/*===========================================================================*\
 *
 *	Routine:	ControllerTerm()
 *
 *	Comments:	Deactivates the controllers
 *
\*===========================================================================*/

static void ControllerTerm(
	void)
{
	int i;

	for (i = 0; i < fNumControllers; i++) {
		if (fControllerRefs[i] != NULL) {
			Q3Controller_Decommission(fControllerRefs[i]);
			/*Q3Object_Dispose(fControllerRefs[i]);*/
			fControllerRefs[i] = NULL;
		}
	}
	fNumControllers = 0;
}


/*===========================================================================*\
 *
 *	Routine:	ControllerFind()
 *
 *	Comments:	Find the index of the controller for the bird with the address
 *				requested, if none exists, returns -1
 *				Assumes the variables, 'fNumBirds and 'fBirdAddresses' are up
 *				to date
 *
\*===========================================================================*/

static short ControllerFind(
	Byte address)
{
	int i;
	short index = -1;

	for (i = 0; i < fNumBirds; i++) {
		if (fBirdAddresses[i] == address) {
			index = i;
			break;
		}
	}
	
	return index;
}


/*===========================================================================*\
 *
 *	Routine:	SerialPortInit()
 *
 *	Comments:	Initialize the serial port for communication with the Flock
 *				on the port requested
 *
\*===========================================================================*/

static OSErr SerialPortInit(
	short port,
	short speed,
	short *serPortIn,
	short *serPortOut)
{
	SerShk flags;
	OSErr err;
	Str255 name;

	*serPortIn = 0;
	*serPortOut = 0;
	
	if (port < 0) { err = paramErr; goto bail; }

	err = GetOutputDriverName(port, name);
	if (err != noErr) goto bail;
	if (DriverIsOpen(name)) { err = openErr; goto bail; }
	err = OpenDriver(name, serPortOut);
	if (err != noErr) goto bail;
	err = GetInputDriverName(port, name);
	if (err != noErr) goto bail;
	if (DriverIsOpen(name)) { err = openErr; goto bail; }
	err = OpenDriver(name, serPortIn);
	if (err != noErr) goto bail;
	
	/* Setup ports communications parameters */
	flags.fXOn = false;		/* output XON/XOFF setting */
	flags.fCTS = false;		/* hardware CTS setting */
	flags.xOn  = 0x11;		/* dc1  ^Q */
	flags.xOff = 0x13;		/* dc3  ^S */
	flags.errs = 0;			/* post no error events */
	flags.evts = 0;			/* post no error events */
	flags.fInX = false;		/* input XON/XOFF setting */
	flags.fDTR = false;		/* hardware DTR setting */
	
	err = SerHShake(*serPortOut, &flags);
	if (err != noErr) goto bail;
	err = SerHShake(*serPortIn, &flags);
	if (err != noErr) goto bail;
	
	/* Flock expects 1 Stopbit, no parity and 8 Databits */
	err = SerReset(*serPortOut, speed|data8|stop10|noParity);
	if (err != noErr) goto bail;
	err = SerReset(*serPortIn, speed|data8|stop10|noParity);
	if (err != noErr) goto bail;

	/* Clear serial receive buffer */
	err = SerRecvFlush(*serPortIn);
	if (err != noErr) goto bail;	

	return err;

bail:
	if (*serPortIn != 0) {
		KillIO(*serPortIn);
		CloseDriver(*serPortIn);
		*serPortIn = 0;
	}
	if (*serPortOut != 0) {
		KillIO(*serPortOut);
		CloseDriver(*serPortOut);
		*serPortOut = 0;
	}
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	SerialPortTerm()
 *
 *	Comments:	Close the serial port drivers.
 *
\*===========================================================================*/

static OSErr SerialPortTerm (
	short serPortIn,
	short serPortOut)
{
	OSErr err, rval;
	
	if (serPortIn != 0) {
		err = KillIO(serPortIn);
		if (err) rval = err;
		err = CloseDriver(serPortIn);
		if (err) rval = err;
	}
	if (serPortOut != 0) {
		err = KillIO(serPortOut);
		if (err) rval = err;
		err = CloseDriver(serPortOut);
		if (err) rval = err;
	}

	return err;
}


#if 0
/*===========================================================================*\
 *
 *	Routine:	EncodeValue()
 *
 *	Comments:	Convert the value to flock format
 *              The value is expected in the -1...1 range
 *
\*===========================================================================*/

static void EncodeValue(
	float value,
	Byte *lsb,
	Byte *msb)
{
	SInt16 word;
	if (value > 0.99996) {
		value = 0.99996;
	}
	if (value < -1) {
		value = -1;
	}
	word = (SInt16)(value*32768);
	*lsb = 0x00FF & word;
	*msb = word >> 8;
}


/*===========================================================================*\
 *
 *	Routine:	EncodeAngle()
 *
 *	Comments:	Convert the angle to flock format
 *              The angle is expected in the -180...180 range
 *
\*===========================================================================*/

static void EncodeAngle(
	float angle,
	Byte *lsb,
	Byte *msb)
{
	EncodeValue(angle/180., lsb, msb);
}


/*===========================================================================*\
 *
 *	Routine:	EncodeQD3DReferenceFrame1()
 *
 *	Comments:	Encode the QD3D reference frame
 *
\*===========================================================================*/

static void EncodeQD3DReferenceFrame1(
	Byte bytes[12])
{
	/* Encode sine and cosine of azimuth (z-angle = -90) */
	EncodeValue(-1., bytes, bytes+1);
	EncodeValue(0., bytes+2, bytes+3);
	/* Encode sine and cosine of elevation (y-angle = 0) */
	EncodeValue(0., bytes+4, bytes+5);
	EncodeValue(1., bytes+6, bytes+7);
	/* Encode sine and cosine of roll (x-angle = -90) */
	EncodeValue(-1., bytes+8, bytes+9);
	EncodeValue(0., bytes+10, bytes+11);
}
#endif


/*===========================================================================*\
 *
 *	Routine:	DecodeValue()
 *
 *	Comments:	Get value in range +.99996=0x7FFF, 0=0x0000, -1=0x8000
 *
\*===========================================================================*/

static void DecodeValue(
	Byte lsb,
	Byte msb,
	float *value)
{
	UInt16 word;
	Byte *bytes = (Byte *) &word;

	bytes[1] = lsb << 1;
	bytes[0] = msb;
	word <<= 1;
	*value = (SInt16) word;
	*value /= (float) 0x8000;
}


/*===========================================================================*\
 *
 *	Routine:	DecodePoint()
 *
 *	Comments:	Get point from the bytes
 *
\*===========================================================================*/

static void DecodePoint(
	Byte bytes[6],
	TQ3Point3D *point)
{
	DecodeValue(bytes[0], bytes[1], &point->x);
	DecodeValue(bytes[2], bytes[3], &point->y);
	DecodeValue(bytes[4], bytes[5], &point->z);
}


/*===========================================================================*\
 *
 *	Routine:	DecodeQuaternion()
 *
 *	Comments:	Get quaternion from the bytes
 *
\*===========================================================================*/

static void DecodeQuaternion(
	Byte bytes[8],
	TQ3Quaternion *quaternion)
{
	DecodeValue(bytes[0], bytes[1], &quaternion->w);
	DecodeValue(bytes[2], bytes[3], &quaternion->x);
	DecodeValue(bytes[4], bytes[5], &quaternion->y);
	DecodeValue(bytes[6], bytes[7], &quaternion->z);
}


/*===========================================================================*\
 *
 *	Routine:	DecodeButtons()
 *
 *	Comments:	Get button states from byte
 *
\*===========================================================================*/

static void DecodeButtons(
	Byte byte,
	unsigned long *buttons)
{
	*buttons = 0x00;
	if ((byte & 0x10) == 0x10) {
		*buttons |= 0x01;
	}
	if ((byte & 0x20) == 0x20) {
		*buttons |= 0x02;
	}
	if ((byte & 0x40) == 0x40) {
		*buttons |= 0x04;
	}
}


/*===========================================================================*\
 *
 *	Routine:	GetBirdError()
 *
 *	Comments:	Get the last error from the bird
 *				Valid connection to bird is expected
 *
\*===========================================================================*/

static OSErr GetBirdError(
	void)
{
	OSErr err;
	Byte command[2];
	
	err = err_BirdNotResponding;
	if (fPort == -1) { goto exit; }

	err = SerRecvFlush(fSerInp);
	if (err != noErr) goto exit;	
	command[0] = 0x4F;	/* Examine value command */
	command[1] = 0x0A;	/* Error code parameter number */
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	err = SerRecvData(command, 1);
	if (err != noErr) goto exit;
	if (command[0] != 0) { err = command[0]; goto exit; }

exit:
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	GetBirdIsStandAlone()
 *
 *	Comments:	Return whether the bird is in standalone mode
 *				Flock should not be running when this routine is called
 *				Valid connection to bird is expected
 *
\*===========================================================================*/

static OSErr GetBirdIsStandAlone(
	Boolean *standAlone)
{
	OSErr err;
	Byte command[5];
	
	*standAlone = true;

	/* Examine the configuration mode, number of FBB devices */
	command[0] = 0x4F;	/* Examine value command */
	command[1] = 0x32;	/* FBB Auto-configuration parameter number */
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	err = SerRecvData(command, 5);
	if (err != noErr) goto exit;
	if (command[0] != 0) {
		*standAlone = false;
	}

exit:	
	return err;
}

/*===========================================================================*\
 *
 *	Routine:	GetBirdsPresent()
 *
 *	Comments:	Return the physical addresses of the birds that are accessible
 *				over the FBB bus and have a receiver
 *				Flock should not be running when this routine is called
 *				Valid connection to bird is expected
 *
\*===========================================================================*/

static OSErr GetBirdsPresent(
	short addresses[MAX_NUM_ADDRESSES],
	short *numAddresses,
	short *numFound,
	Boolean standAlone)
{
	OSErr err;
	int i;
	Byte command[13];
	
	*numAddresses = 0;
	*numFound = 0;
	
	err = noErr;
	if (standAlone) {
		*numAddresses = 1;
		*numFound = 1;
		addresses[0] = 0;
	}
	else {
		/* Check flock system status */
		command[0] = 0x4F;	/* Examine value command */
		command[1] = 0x24;	/* Flock System Status parameter number */
		err = SerSendData(command, 2);
		if (err != noErr) goto exit;
		err = SerRecvData(command, 14);
		if (err != noErr) goto exit;

		for (i = 0; i < 14; i++) {
			/* For each byte bit 5 tells us that the bird at address i has a
			 * receiver. Bit 7 flags whether the device at address i is accessible
			 * over the FBB bus.
			 */
			if (command[i] & 0x20) {
				(*numFound)++;
				if (command[i] & 0x80) {
					addresses[*numAddresses] = i+1;
					(*numAddresses)++;
				}
			}
		}
	}
		
exit:	
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	SetBirdsRunning()
 *
 *	Comments:	Use the FBB Auto-Configuration command to start the birds at
 *				the addresses given.
 *				Flock should not be running when this routine is called
 *				Valid connection to bird is expected
 *
\*===========================================================================*/

static OSErr SetBirdsRunning(
	short addresses[MAX_NUM_ADDRESSES],
	short numAddresses,
	Boolean standAlone)
{
	OSErr err = noErr;
	int i;
	Byte command[3];
		
	/* Send auto configuration only when not in stand alone mode */
	if (!standAlone) {
		/* Check whether the first address == 1 and the rest is contiguous */
		for (i = 0; i < numAddresses; i++) {
			if (addresses[i] != i + 1) {
				err = err_AddressConfiguration;
				goto exit;
			}
		}

		Wait(300);
		command[0] = 0x50;				/* Change value command */
		command[1] = 0x32;				/* FBB Auto-configuration parameter number */
		command[2] = numAddresses;		/* Number of Bird electronic units */
		err = SerSendData(command, 3);
		if (err != noErr) goto exit;
		Wait(300);
		if ((err = GetBirdError()) != noErr) goto exit;
	}

exit:	
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	UpdateControllerRelative()
 *
 *	Comments:	Update the QD3D controller from the previous and new pos/quat
 *
\*===========================================================================*/

static void UpdateControllerRelative(
	int index,
	TQ3Point3D *prevP,
	TQ3Quaternion *prevQ,
	TQ3Point3D *currP,
	TQ3Quaternion *currQ,
	unsigned long buttons)
{
	TQ3Vector3D offset, qd3dOffset;
	TQ3Quaternion rotation, qd3dRotation;
	TQ3Boolean track2DCursor;

	Q3Controller_Track2DCursor(fControllerRefs[index], &track2DCursor);
	if ((track2DCursor == kQ3False) ||
	    ((track2DCursor == kQ3True) && fTrack2DCursor)) {
		Q3Point3D_Subtract(currP, prevP, &offset);
		qd3dOffset.x = -fScaleX * offset.y;
		qd3dOffset.y = -fScaleY * offset.z;
		qd3dOffset.z =  fScaleZ * offset.x;
		Q3Controller_MoveTrackerPosition(fControllerRefs[index], &qd3dOffset);
		
		Q3Quaternion_Invert(prevQ, &rotation);
		Q3Quaternion_Multiply(&rotation, currQ, &rotation);
		Q3Quaternion_Normalize(&rotation, &rotation);
		qd3dRotation.w = rotation.w;
		qd3dRotation.x = rotation.y;
		qd3dRotation.y = rotation.z;
		qd3dRotation.z = -rotation.x;
		Q3Controller_MoveTrackerOrientation(fControllerRefs[index], &qd3dRotation);

		Q3Controller_SetButtons(fControllerRefs[index], buttons);
	}
}


/*===========================================================================*\
 *
 *	Routine:	UpdateControllerAbsolute()
 *
 *	Comments:	Update the QD3D controller from the new pos/quat
 *
\*===========================================================================*/

static void UpdateControllerAbsolute(
	int index,
	TQ3Point3D *point,
	TQ3Quaternion *quaternion,
	unsigned long buttons)
{
	TQ3Boolean track2DCursor;
	TQ3Point3D qd3dPoint;
	TQ3Quaternion qd3dQuaternion;
	
	Q3Controller_Track2DCursor(fControllerRefs[index], &track2DCursor);
	if ((track2DCursor == kQ3False) ||
	    ((track2DCursor == kQ3True) && fTrack2DCursor)) {
		qd3dPoint.x = -fScaleX * point->y;
		qd3dPoint.y = -fScaleY * point->z;
		qd3dPoint.z =  fScaleZ * point->x;
		Q3Controller_SetTrackerPosition(fControllerRefs[index], &qd3dPoint);
		qd3dQuaternion.w = quaternion->w;
		qd3dQuaternion.x = quaternion->y;
		qd3dQuaternion.y = quaternion->z;
		qd3dQuaternion.z = -quaternion->x;
		Q3Controller_SetTrackerOrientation(fControllerRefs[index], &qd3dQuaternion);
		Q3Controller_SetButtons(fControllerRefs[index], buttons);
	}
}


/******************************************************************************
 **	Public Routines
 *****************************************************************************/

/*===========================================================================*\
 *
 *	Routine:	BirdInit()
 *
 *	Comments:	Initialize QD3D controller, serial port and the bird
 *
\*===========================================================================*/

OSErr BirdInit(
	short portID,
	short speed,
	TBirdConfiguration *config,
#if OPERATION_MODE
	TBirdOperation operation,
#endif
	TBirdMode mode,
	float inch2qd3dX,
	float inch2qd3dY,
	float inch2qd3dZ,
	Boolean controlCursor)
{
	OSErr err;

	err = BirdSetPort(portID, speed);
	if (err) goto exit;
	err = BirdSetConfiguration(config);
	if (err) goto exit;
#if OPERATION_MODE
	err = BirdSetOperation(operation);
	if (err) goto exit;
#endif
	BirdSetMode(mode);
	BirdSetInch2QD3D(inch2qd3dX, inch2qd3dY, inch2qd3dZ);
	BirdSetControlCursor(controlCursor);
	err = BirdStartRunning();
	if (err) goto exit;
	
exit:
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	BirdTerm()
 *
 *	Comments:	Close QD3D controller, serial port and the bird
 *
\*===========================================================================*/

OSErr BirdTerm(
	void)
{
	OSErr err;
	
	err = BirdStopRunning();
	err = BirdSetPort(-1, baud9600);
	ControllerTerm();
	
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	BirdSetPort()
 *
 *	Comments:	Close current serial port and set new port and speed
 *				Will stop running the flock if it was running
 *
\*===========================================================================*/

OSErr BirdSetPort(
	short port,
	short speed)
{
	OSErr err;
	
	err = noErr;
	if ((port == fPort) && (speed == fPortSpeed)) goto exit;
	err = BirdStopRunning();
	if (err != noErr) goto exit;

	/* Close the current port */
	if (fPort != -1) {
		err = SerialPortTerm(fSerInp, fSerOut);
		fSerInp = 0;
		fSerOut = 0;
		fPort = -1;
		if (err != noErr) goto exit;
	}
	/* Open the new port */
	if (port == -1) goto exit;
	err = SerialPortInit(port, speed, &fSerInp, &fSerOut);
	if (err != noErr) goto exit;
	fPort = port;
	fPortSpeed = speed;

exit:
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	BirdGetPort()
 *
 *	Comments:	Return the current serial port
 *
\*===========================================================================*/

short BirdGetPort(
	 void)
{
	return fPort;
}


/*===========================================================================*\
 *
 *	Routine:	BirdGetPortSpeed()
 *
 *	Comments:	Return the current transmission speed
 *
\*===========================================================================*/

short BirdGetPortSpeed(
	 void)
{
	return fPortSpeed;
}


/*===========================================================================*\
 *
 *	Routine:	BirdGetInformation()
 *
 *	Comments:	Return information about the flock hardware
 *				This will stop running the flock
 *
\*===========================================================================*/

OSErr BirdGetInformation(
	TBirdInformation *info)
{
	OSErr err;
	Byte command[5];
	short addresses[MAX_NUM_ADDRESSES];
	
	err = err_BirdNotResponding;
	if (fPort == -1) { goto exit; }
	err = BirdStopRunning();
	if (err != noErr) goto exit;

	/* Get the stand-alone state and number of birds */
	GetBirdIsStandAlone(&info->standAlone);
	GetBirdsPresent(addresses,
	                &info->numBirdsOperating,
	                &info->numBirdsFound,
	                info->standAlone);
	
	/* Examine the software revision */
	command[0] = 0x4F;	/* Examine value command */
	command[1] = 0x01;	/* Software revision number parameter number */
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	err = SerRecvData(command, 2);
	if (err != noErr) goto exit;
	info->softwareRevInt = command[0];
	info->softwareRevFrac = command[1];

	/* Examine the crystal speed */
	command[0] = 0x4F;	/* Examine value command */
	command[1] = 0x02;	/* Crystal speed parameter number */
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	err = SerRecvData(command, 2);
	if (err != noErr) goto exit;
	info->crystalSpeed = command[0];

exit:	
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	BirdGetConfiguration()
 *
 *	Comments:	Return the current configuration
 *				This will stop running the flock
 *
\*===========================================================================*/

OSErr BirdGetConfiguration(
	TBirdConfiguration *config)
{
	OSErr err;
	Byte command[5];
	
	err = err_BirdNotResponding;
	if (fPort == -1) { goto exit; }

	err = BirdStopRunning();
	if (err != noErr) goto exit;
	
	/* Examine bird status */
	command[0] = 0x4F;	/* Examine value command */
	command[1] = 0x00;	/* Bird status parameter number */
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	err = SerRecvData(command, 2);
	if (err != noErr) goto exit;
#if 0
	/* Would like to use this but the sync bits are never set in my FOB
	 * so we use the saved file scope variable
	 */
	config->sync = sync_None;
	if (command[1] & 0x02) {
		/* We know it is in CRT sync mode, don't know which, set CRT2 */
		config->sync = sync_CRT2;
	}
	else if (command[1] & 0x08) {
		config->sync = sync_Host;
	}
#else
	config->sync = fBirdSync;
#endif

	/* Examine the filter on/off status */
	command[0] = 0x4F;	/* Examine value command */
	command[1] = 0x04;	/* Filter on/off status parameter number */
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	err = SerRecvData(command, 2);
	if (err != noErr) goto exit;
	config->ACNarrowOn = command[0] & 1<<2 ? false : true;
	config->ACWideOn   = command[0] & 1<<1 ? false : true;
	config->DCOn       = command[0] & 1<<0 ? false : true;

	/* Examine the measurement rate */
	command[0] = 0x4F;	/* Examine value command */
	command[1] = 0x07;	/* Measurement rate parameter number */
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	err = SerRecvData(command, 2);
	if (err != noErr) goto exit;
	config->measurementRate = command[0];
	config->measurementRate = command[1] << 8;

	/* Examine the sudden output change elimination */
	command[0] = 0x4F;	/* Examine value command */
	command[1] = 0x0E;	/* Sudden output change elimination parameter number */
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	err = SerRecvData(command, 1);
	if (err != noErr) goto exit;
	config->suddenOutputElimination = command[0];
	
	/* Examine the range setting */
	command[0] = 0x4F;	/* Examine value command */
	command[1] = 0x03;	/* Position scaling parameter number */
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	err = SerRecvData(command, 2);
	if (err != noErr) goto exit;
	config->range = command[0] == 0 ? range_36 : range_72;

exit:	
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	BirdSetConfiguration()
 *
 *	Comments:	Set new Flock configuration
 *				This will stop running the flock
 *
\*===========================================================================*/

OSErr BirdSetConfiguration(
	TBirdConfiguration *config)
{
	OSErr err;
	Byte command[13];
	int i;
	short numFound;
	float voltage, scanRate;
	
	err = err_BirdNotResponding;
	if (fPort == -1) { goto exit; }	
	err = BirdStopRunning();
	if (err != noErr) goto exit;

	/* Update the number of birds in the flock and the stand-alone state */
	err = GetBirdIsStandAlone(&fStandAlone);
	if (err != noErr) goto exit;
	err = GetBirdsPresent(fBirdAddresses, &fNumBirds, &numFound, fStandAlone);
	if (err != noErr) goto exit;
	if (fNumBirds < 1) {
		err = err_BirdNotResponding;
		goto exit;
	}
	/* Start running the birds in the flock if not in stand-alone mode */
	err = SetBirdsRunning(fBirdAddresses, fNumBirds, fStandAlone);
	if (err != noErr) goto exit;

	if ((config->sync == sync_CRT1) || (config->sync == sync_CRT2)) {
		/* Check if the SYNC pick up produces a valid signal */
		config->sync = sync_None;
		err = BirdGetSyncData(&voltage, &scanRate);
		if (err != noErr) goto exit;
		if (voltage > 1) {
			if ((scanRate > 72) && (scanRate < 144)) { 
				config->sync = sync_CRT1;
			}
			else if ((scanRate > 50) && (scanRate <= 72)) {
				config->sync = sync_CRT2;
			}
		}
	}
	/* Set SYNC settings for master bird only */
	command[0] = 0x41;
	command[1] = config->sync;
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	err = GetBirdError();
	if (err != noErr) goto exit;
	/* Would not have to use this if the Bird Status examine would work !!! */
	fBirdSync = config->sync;

	/* Set the measurement rate for the master bird only  */
	if (config->sync == sync_None) {
		if (config->measurementRate > 36864) {
			config->measurementRate = 36864;
		}
		command[0] = 0x50;	/* Change value command */
		command[1] = 0x07;	/* Measurement rate parameter number */
		command[2] = config->measurementRate & 0x00FF;
		command[3] = config->measurementRate >> 8;
		err = SerSendData(command, 4);
		if (err != noErr) goto exit;
		if ((err = GetBirdError()) != noErr) goto exit;
	}		
	
	/* For each of the birds in the flock, set filter settings, and sudden
	 * output elimination
	 */
	for (i = 0; i < fNumBirds; i++) {
		/* Set filter on/off settings */
		if (!fStandAlone) {
			/* Send RS232 to FBB command to master to address the right bird */
			command[0] = 0xF0 + fBirdAddresses[i];
			err = SerSendData(command, 1);
			if (err != noErr) goto exit;
		}
		command[0] = 0x50;				/* Change value command */
		command[1] = 0x04;				/* FBB filter on/off parameter number */
		command[2] = 0x00;
		if (!config->ACNarrowOn) command[2] |= 1<<2;	
		if (!config->ACWideOn)   command[2] |= 1<<1;	
		if (!config->DCOn)       command[2] |= 1<<0;
		command[3] = 0x00;
		err = SerSendData(command, 4);
		if (err != noErr) goto exit;
		if ((err = GetBirdError()) != noErr) goto exit;
		
		/* Set the sudden output elimination */
		if (!fStandAlone) {
			/* Send RS232 to FBB command to master to address the right bird */
			command[0] = 0xF0 + fBirdAddresses[i];
			err = SerSendData(command, 1);
			if (err != noErr) goto exit;
		}
		command[0] = 0x50;	/* Change value command */
		command[1] =  0x0E;	/* Sudden output change elimination parameter number */
		command[2] = config->suddenOutputElimination;
		err = SerSendData(command, 3);
		if (err != noErr) goto exit;
		if ((err = GetBirdError()) != noErr) goto exit;

		/* Set the range */
		if (!fStandAlone) {
			/* Send RS232 to FBB command to master to address the right bird */
			command[0] = 0xF0 + fBirdAddresses[i];
			err = SerSendData(command, 1);
			if (err != noErr) goto exit;
		}
		command[0] = 0x50;	/* Change value command */
		command[1] = 0x03;	/* Position scaling parameter number */
		command[2] = config->range == range_36 ? 0x00 : 0x01;
		command[3] = 0x00;
		err = SerSendData(command, 4);
		if (err != noErr) goto exit;
		if ((err = GetBirdError()) != noErr) goto exit;
	}
	
	fBirdRange = config->range;
	if (fBirdRange == range_36) {
		fScaleX = 36 * fInch2Qd3dX;
		fScaleY = 36 * fInch2Qd3dY;
		fScaleZ = 36 * fInch2Qd3dZ;
	}
	else {
		fScaleX = 72 * fInch2Qd3dX;
		fScaleY = 72 * fInch2Qd3dY;
		fScaleZ = 72 * fInch2Qd3dZ;
	}	
		
exit:
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	BirdGetSyncData()
 *
 *	Comments:	Get the sync data
 *				This will stop the flock if it was running. Before continuing
 *				the flock should be set in a real sync mode with BirdSetSync!
 *
\*===========================================================================*/

OSErr BirdGetSyncData(
	float *outVoltage,
	float *outScanRate)
{
	OSErr err;
	Byte command[4];
	UInt16 value;
	float clock;
	
	err = err_BirdNotResponding;
	if (fPort == -1) { goto exit; }
	err = BirdStopRunning();
	if (err != noErr) goto exit;

	/* Send SYNC command */
	command[0] = 0x41;
	command[1] = 0xFF;	/* CRTSYNC 255, pickup placement */
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	/* Get the data from the port and decode */
	err = SerRecvData(command, 4);
	if (err != noErr) goto exit;
	DecodeValue(command[0], command[1], outVoltage);
	*outVoltage *= 5.;
	value = command[3] << 8;
	value |= command[2];
	/* Examine the crystal speed */
	command[0] = 0x4F;	/* Examine value command */
	command[1] = 0x02;	/* Crystal speed parameter number */
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;
	err = SerRecvData(command, 2);
	if (err != noErr) goto exit;
	clock = 8. / ((float)command[0]);
	if (value == 0) {
		*outScanRate = 0.;
	}
	else {
		*outScanRate = 500000. / (clock * ((float)value));
	}

	/* TEMP!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
	command[0] = 0x41;
	command[1] = sync_None;
	err = SerSendData(command, 2);
	if (err != noErr) goto exit;

exit:
	return err;
}


#if OPERATION_MODE
/*===========================================================================*\
 *
 *	Routine:	BirdGetOperation()
 *
 *	Comments:	Get the current operation mode of the bird driver
 *
\*===========================================================================*/

TBirdOperation
BirdGetOperation(
	void)
{
	return fOperation;
}


/*===========================================================================*\
 *
 *	Routine:	BirdSetOperation()
 *
 *	Comments:	Set the current operation mode of the bird driver
 *				This will stop running the flock
 *
\*===========================================================================*/

OSErr
BirdSetOperation(
	TBirdOperation operation)
{
	OSErr err;
	
	err= noErr;
	if (operation == fOperation) goto exit;
	err = BirdStopRunning();
	if (err != noErr) goto exit;
	fOperation = operation;

exit:
	return err;
}
#endif


/*===========================================================================*\
 *
 *	Routine:	BirdGetMode()
 *
 *	Comments:	Get the current measure mode of the bird driver
 *
\*===========================================================================*/

TBirdMode
BirdGetMode(
	void)
{
	return fMode;
}


/*===========================================================================*\
 *
 *	Routine:	BirdSetMode()
 *
 *	Comments:	Set the current measure mode of the bird driver
 *
\*===========================================================================*/

void BirdSetMode(
	TBirdMode mode)
{
	if (fMode == mode) return;
	fMode = mode;
	fSkip = 0;
}



/*===========================================================================*\
 *
 *	Routine:	BirdGetIsRunning()
 *
 *	Comments:	Return whether the flock is running
 *
\*===========================================================================*/

Boolean BirdGetIsRunning(
	void)
{
	return fIsRunning;
}


/*===========================================================================*\
 *
 *	Routine:	BirdStopRunning()
 *
 *	Comments:	Stop reading location/orientation data from the port
 *
\*===========================================================================*/

OSErr BirdStopRunning(
	void)
{
	OSErr err;
	Byte command[3];
	
	if (!fIsRunning) { return noErr; }
	err = err_BirdNotResponding;
	if (fPort == -1) { goto exit; }
	err = noErr;
#if OPERATION_MODE
	if (fMode == mode_Poll) {
		if (fSerInp) {
			err = SerRecvFlush(fSerInp);
		}
	}
	else {
#endif
		/* Stop the data stream from the flock by sending point command */
		command[0] = 0x42;
		err = SerSendData(command, 1);

		/* Kill pending requests for the serial port */
		PBKillIOAsync((ParmBlkPtr) &fSerParm);
		SerRecvFlush(fSerInp);
		DisposeRoutineDescriptor(fDefTask.dtAddr);
		DisposeRoutineDescriptor(fSerParm.ioCompletion);
		Wait(300);

		/* Set group mode off */
		command[0] = 0x50;	/* Change value command */
		command[1] = 0x23;	/* Group mode parameter number */
		command[2] = 0x00;
		err = SerSendData(command, 3);
		if (err != noErr) goto exit;
		if ((err = GetBirdError()) != noErr) goto exit;
#if OPERATION_MODE
	}
#endif
	fIsRunning = false;

exit:
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	BirdStartRunning()
 *
 *	Comments:	Start reading location/orientation data from the port
 *
\*===========================================================================*/

OSErr BirdStartRunning(
	void)
{
	OSErr err;
	int i;
	Byte command[3];
	
	if (fIsRunning) { return noErr; }
	err = err_BirdNotResponding;
	if (fPort == -1) { goto exit; }

	/* Get the addresses of the birds */
//	err = GetBirdAddresses(fBirdAddresses, &fNumBirds, &fStandAlone);
//	if (err != noErr) goto exit;
	if (fNumBirds != fNumControllers) {
		ControllerTerm();
		err = ControllerInit(fNumBirds);
		if (err != noErr) goto exit;
	}
	
	/* Set all birds in position/quaternion mode and have the buttons reported
	 * in a byte appended to every data record
	 */
	err = SerRecvFlush(fSerInp);
	if (err != noErr) goto exit;
	for (i = 0; i < fNumBirds; i++) {
		if (!fStandAlone) {
			/* Send address of bird by sending RS232 to FBB command */
			command[0] = 0xF0 + fBirdAddresses[i];
			err = SerSendData(command, 1);
			if (err != noErr) goto exit;
		}
		/* Issue a position/quaternion command */
		command[0] = 0x5D;
		err = SerSendData(command, 1);
		if (err != noErr) goto exit;
		if ((err = GetBirdError()) != noErr) goto exit;
		if (!fStandAlone) {
			/* Send address of bird by sending RS232 to FBB command */
			command[0] = 0xF0 + fBirdAddresses[i];
			err = SerSendData(command, 1);
			if (err != noErr) goto exit;
		}
		/* Issue a button mode command */
		command[0] = 0x4D;
		command[1] = 0x01;
		err = SerSendData(command, 2);
		if (err != noErr) goto exit;
		if ((err = GetBirdError()) != noErr) goto exit;
	}

	/* Set the right group mode */
#if OPERATION_MODE
	if ((fOperation == mode_Poll) || (fStandAlone)) {
#else
	if (fStandAlone) {
#endif
		/* Turn group mode off for all the birds */
		/* IS THIS NECESSARY???????????????????? */
		for (i = 0; i < fNumBirds; i++) {
			if (!fStandAlone) {
				/* Send address of bird by sending RS232 to FBB command */
				command[0] = 0xF0 + fBirdAddresses[i];
				err = SerSendData(command, 1);
				if (err != noErr) goto exit;
			}
			command[0] = 0x50;	/* Change value command */
			command[1] = 0x23;	/* Group mode parameter number */
			command[2] = 0x00;	/* Do not use group mode */
			err = SerSendData(command, 3);
			if (err != noErr) goto exit;
			if ((err = GetBirdError()) != noErr) goto exit;
		}
		fPacketSize = PACKET_SIZE_STAND_ALONE;
	}
	else {
		/* Turn group mode on for all the birds */
		/* IS THIS NECESSARY???????????????????? */
		for (i = 0; i < fNumBirds; i++) {
			if (!fStandAlone) {
				/* Send RS232 to FBB command to master to address the right bird */
				command[0] = 0xF0 + fBirdAddresses[i];
				err = SerSendData(command, 1);
				if (err != noErr) goto exit;
			}
			command[0] = 0x50;	/* Change value command */
			command[1] = 0x23;	/* Group mode parameter number */
			command[2] = 0x01;	/* Use group mode */
			err = SerSendData(command, 3);
			if (err != noErr) goto exit;
			if ((err = GetBirdError()) != noErr) goto exit;
		}
		fPacketSize = PACKET_SIZE_MASTER_SLAVE;
	}
	
	/* If stream mode, start the stream */
#if OPERATION_MODE
	if (fOperation == mode_Stream) {
#endif
		/* Prepare the flock for stream mode */
		fBufferCount = 0;
		/* Set up the deferred task block */
		fDefTask.qType		= dtQType;
		fDefTask.dtAddr		= NewDeferredTaskProc(DeferredCompletion);
		fDefTask.dtParam	= 0;
		fDefTask.dtReserved	= 0;
		
		/* Set up the serial parameter block */
		fSerParm.ioCompletion	= NewIOCompletionProc(ASyncCompletion);
		fSerParm.ioRefNum		= fSerInp;
		fSerParm.ioPosMode		= fsAtMark;
		fSerParm.ioPosOffset	= 0;
		fSerParm.ioBuffer		= (Ptr) fBuffer;
		fSerParm.ioReqCount		= fPacketSize;

		/* Start the read chain by reading the first character */
		err = SerRecvFlush(fSerInp);
		err = PBReadAsync((ParmBlkPtr) &fSerParm);
		
		/* Start the data stream from flock by issuing a stream command */
		command[0] = 0x40;
		err = SerSendData(command, 1);
		if (err != noErr) {
			PBKillIOAsync((ParmBlkPtr) &fSerParm);
			goto exit;
		}
#if OPERATION_MODE
	}
#endif
	fSkip = 0;
	fIsRunning = true;

exit:
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	BirdPoll()
 *
 *	Comments:	Poll the flock for new coordinates
 *
\*===========================================================================*/

#if OPERATION_MODE
OSErr BirdPoll(
	void)
{
	OSErr err;
	Byte command[15];
	TQ3Point3D point;
	TQ3Quaternion quaternion;
	unsigned long buttons;
	int i;

	if ((!fIsRunning) || (fOperation != mode_Poll)) { return noErr; }

	/* Check if we're connected to a port */
	err = err_BirdNotResponding;
	if (fPort == -1) { goto exit; }

	for (i = 0; i < fNumControllers; i++) {
		if (!fStandAlone) {
			/* Send RS232 to FBB command to master to address the right bird */
			command[0] = 0xF0 + fBirdAddresses[i];
			err = SerSendData(command, 1);
			if (err != noErr) goto exit;
		}
		/* Issue a point command to receive new data from bird */
		command[0] = 0x42;
		err = SerSendData(command, 1);
		if (err != noErr) goto exit;
		/* Get the data */
		err = SerRecvData(command, 15);
		if (err != noErr) goto exit;
		command[0] &= 0x7F;	/* Clear high/phasing bit */
		DecodePoint(command, &point);
		DecodeQuaternion(command+6, &quaternion);
		Q3Quaternion_Normalize(&quaternion, &quaternion);
		DecodeButtons(command[14], &buttons);
		if (fSkip >= fNumSkip) {
			if (fMode == mode_Relative) {
				/* Relative update, update if this not the first measurement */		
				UpdateControllerRelative(i,
				                         &fPrevPoint[i],
				                         &fPrevQuaternion[i],
				                         &point,
				                         &quaternion,
				                         buttons);
			}
			else {
				/* Absolute update */
				UpdateControllerAbsolute(i, &point, &quaternion, buttons);
			}
		}
		else {
			fSkip++;
		}
		/* Save the current measurement */
		fPrevPoint[i] = point;
		fPrevQuaternion[i] = quaternion;
	}
		
exit:
	return err;
}
#endif

/*===========================================================================*\
 *
 *	Routine:	BirdGetInch2QD3D()
 *
 *	Comments:	Return the factors that control the bird coordinate to QD3D 
 *				tracker coordinate conversion 
 *
\*===========================================================================*/

void BirdGetInch2QD3D(
	float *x,
	float *y,
	float *z)
{
	*x = fInch2Qd3dX;
	*y = fInch2Qd3dY;
	*z = fInch2Qd3dZ;
}


/*===========================================================================*\
 *
 *	Routine:	BirdSetInch2QD3D()
 *
 *	Comments:	Set the factors that control the bird coordinate to QD3D 
 *				tracker coordinate conversion 
 *
\*===========================================================================*/

void BirdSetInch2QD3D(
	float x,
	float y,
	float z)
{
	fInch2Qd3dX = x;
	fInch2Qd3dY = y;
	fInch2Qd3dZ = z;
	if (fBirdRange == range_36) {
		fScaleX = 36 * fInch2Qd3dX;
		fScaleY = 36 * fInch2Qd3dY;
		fScaleZ = 36 * fInch2Qd3dZ;
	}
	else {
		fScaleX = 72 * fInch2Qd3dX;
		fScaleY = 72 * fInch2Qd3dY;
		fScaleZ = 72 * fInch2Qd3dZ;
	}	
}


/*===========================================================================*\
 *
 *	Routine:	BirdSetControlCursor()
 *
 *	Comments:	Set the boolean that controls the bird to system cursor connec-
 *				tion
 *
\*===========================================================================*/

void BirdSetControlCursor(
	Boolean controlCursor)
{
	fTrack2DCursor = controlCursor;
}


/*===========================================================================*\
 *
 *	Routine:	BirdGetControlCursor()
 *
 *	Comments:	Get the boolean that controls the bird to system cursor connec-
 *				tion
 *
\*===========================================================================*/

Boolean BirdGetControlCursor(
	void)
{
	return fTrack2DCursor;
}