/******************************************************************************
 **																			 **
 ** 	Module:		DriverShell.c											 **
 ** 																		 **
 ** 																		 **
 ** 	Purpose: 	Driver shell for Ascension Flock Of Birds				 **
 **					Runs as a Macintosh application, mainly so that it can	 **
 ** 				have a Quit menu item.									 **
 **					Adapted for Flock Of Birds by Maarten Gribnau			 **
 **					email: m.w.gribnau@io.tudelft.nl						 **
 **					Based on code for Magellan driver.						 **
 ** 				Copyright (C) 1992-96 Apple Computer, Inc.				 **
 **					All rights reserved.									 **
 ** 																		 **
 ** 																		 **
 *****************************************************************************/

#include <Controls.h>
#include <Devices.h>
#include <Dialogs.h>
#include <Fonts.h>
#include <Processes.h>
#include <Sound.h>
#include <Strings.h>
#include <ToolUtils.h>

#include <QD3D.h>
#include <QD3DErrors.h>
#include <stdio.h>
#include <string.h>

#include "DriverBird.h"
#include "PrefFile.h"
#include "SerialDriverNames.h"
#include "DriverErrors.h"


// Global variable
TPreferences gPreferences;
FSSpec gPrefFile;
const OSType  Creator_Preferences	= 'MG16';

/******************************************************************************
 **																			 **
 **									Constants								 **
 **																			 **
 *****************************************************************************/

enum {									/* MENU resource IDs */
	menuApple = 128,
	menuFile,
	menuDriver = 256
};


enum {									/* Apple menu items */
	appleAbout = 1
};


enum {									/* File menu items */
	fileSerialPort = 1,
	fileFlockConfig,
	fileQD3DConfig,
	fileStartStop,
	fileQuit
};


enum {									/* Resource IDs */
	alrtQD3DError = 255,
	alrtQD3DWarning = 256,
	alrtQD3DNotice = 257,
	alrtOSError = 1001,
	alrtBirdError = 1002,
	alrtDriverError = 1003
};

enum {
	dlgAbout = 1000,
	dlgSerialPort = 2000,
	dlgFlockConfig = 2001,
	dlgQD3DConfig
};

enum {
	butSerialPortOk = 1,
	butSerialPortCancel,
	menuPort = 3,
	but2400 = 4,
	but4800,
	but9600,
	but19200,
	but57600
};

enum {
	butFlockConfigOk = 1,
	butFlockConfigCancel,
	radioACNarrow = 4,
	radioACWide,
	checkDC = 6,
	checkSuddenOutputElimination = 7,
	checkSyncToMonitor = 8,
	editMeasurementRate = 10,
	statVoltage = 13,
	statScanRate = 16,
	statNumBirdsOperating = 19,
	statSoftwareRev = 21,
	statCrystalSpeed = 23,
	statNumBirdsFound = 26,
	radioRange36 = 28,
	radioRange72 = 29
};


enum {
	butQD3DConfigOk = 1,
	butQD3DConfigCancel,
	editInch2qd3dX = 3,
	editInch2qd3dY,
	editInch2qd3dZ,
//	radioPoll = 10,
//	radioStream = 11,
	radioRelative = 10,
	radioAbsolute = 11,
	checkControlCursor = 12
};

enum {
	strxMenuItems = 128
};
enum {
	strStop = 1,
	strStart
};
enum {
	butQD3DOk = 1,
	butQD3DQuit = 2
};

/******************************************************************************
 **																			 **
 **							Variables and Forward Declarations				 **
 **																			 **
 *****************************************************************************/

static MenuHandle		fMenuApple;
static MenuHandle		fMenuFile;
static EventRecord		fEvent;
static Boolean			fExit;
static TQ3Error			fFirstError		= kQ3ErrorNone;
static TQ3Error			fLastError		= kQ3ErrorNone;
static unsigned long	fNumErrors		= 0;
static TQ3Warning		fFirstWarning	= kQ3WarningNone;
static TQ3Warning		fLastWarning	= kQ3WarningNone;
static unsigned long	fNumWarnings	= 0;
static TQ3Notice		fFirstNotice	= kQ3NoticeNone;
static TQ3Notice		fLastNotice		= kQ3NoticeNone;
static unsigned long	fNumNotices		= 0;


static pascal Boolean OkCancelDlgFilter(
	DialogPtr theDlg,
	EventRecord *theEvent,
	short *itemHit);

static void FlipItem(
	DialogPtr dlg,
	short item);

static void SetTextItemInt(
	DialogPtr dlg,
	short item,
	int value);

static void GetTextItemInt(
	DialogPtr dlg,
	short item,
	int *value);

static void SetTextItemFloat(
	DialogPtr dlg,
	short item,
	float value);

static void GetTextItemFloat(
	DialogPtr dlg,
	short item,
	float *value);
	
static void UpdateStartStop(
	void);

static void BuildDriverMenu(
	MenuHandle menu);

static void ClickInWindow(
	WindowPtr wind,
	Point where);

static void	ErrorAlert(
	OSErr err);

static void DoAppleAbout(
	void);

static void DoAppleDeskAcc(
	int item);

void DoFileSerialPort(
	void);

void DoFlockConfig(
	void);

void DoQD3DConfig(
	void);

static void DoFileQuit(
	void);

static void DoMenuCommand(
	long itemVal);

static void WindowDrag(
	WindowPtr wind,
	Point where);

static void DoEventMouseDown(
	void);

static void DoEventKeyDown(
	void);

static void DoEventUpdate(
	WindowPtr wind);

static void DoEventActivate(
	WindowPtr wind);

static void Init(
	void);

static void Loop(	
	void);

static void Term(
	void);

#if 0
OSErr DoSetPort(
	short port,
	short speed);
#endif

void ShowError();
void ShowWarning();
void ShowNotice();
void ErrorMethod(
	TQ3Error firstError,
	TQ3Error lastError,
	long reference);
void WarningMethod(
	TQ3Warning firstWarning,
	TQ3Warning lastWarning,
	long reference);
void NoticeMethod(
	TQ3Notice firstNotice,
	TQ3Notice lastNotice,
	long reference);

void main(
	void);



/******************************************************************************
 **																			 **
 **								Basic Mac Glue Routines						 **
 **																			 **
 *****************************************************************************/


/*===========================================================================*\
 *
 *	Routine:	OkCancelDlgFilter()
 *
 *	Comments:	Dlg filter for default buttons Ok and Cancel
 *
\*===========================================================================*/

pascal Boolean OkCancelDlgFilter(
	DialogPtr /*theDlg*/,
	EventRecord *theEvent,
	short *itemHit)
{
	if (theEvent->what == keyDown) {
		switch ((theEvent->message) & charCodeMask) {
			case 0x0D:
			case 0x03:
				*itemHit = 1;
				return true;
			case 0x1B:
				*itemHit = 2;
				return true;
		}
	}
	return false;
}


/*===========================================================================*\
 *
 *	Routine:	FlipItem()
 *
 *	Comments:	Useful little subroutine for inverting dialog button values.
 *
\*===========================================================================*/

void FlipItem(
	DialogPtr dlg,
	short item)
{
	short         iType;
	Rect          iRect;
	ControlHandle iHandle;

	GetDialogItem(dlg, item, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, !GetControlValue(iHandle));
}


/*===========================================================================*\
 *
 *	Routine:	SetTextItemInt()
 *
 *	Comments:	Useful little subroutine for setting dialog text edit values.
 *
\*===========================================================================*/

void SetTextItemInt(
	DialogPtr dlg,
	short item,
	int value)
{
	short         iType;
	Rect          iRect;
	ControlHandle iHandle;
	Str255        text;

	GetDialogItem(dlg, item, &iType, (Handle *)&iHandle, &iRect);
	sprintf(((char *)text)+1, "%d", value);
	text[0] = strlen(((char *)text)+1);
	SetDialogItemText((Handle)iHandle, text);
}


/*===========================================================================*\
 *
 *	Routine:	GetTextItemInt()
 *
 *	Comments:	Useful little subroutine for getting dialog text edit values.
 *
\*===========================================================================*/

void GetTextItemInt(
	DialogPtr dlg,
	short item,
	int *value)
{
	short         iType;
	Rect          iRect;
	ControlHandle iHandle;
	Str255        text;

	GetDialogItem(dlg, item, &iType, (Handle *)&iHandle, &iRect);
	GetDialogItemText((Handle)iHandle, text);
	p2cstr(text);
	sscanf((char *)text, "%d", value);
}


/*===========================================================================*\
 *
 *	Routine:	SetTextItemFloat()
 *
 *	Comments:	Useful little subroutine for setting dialog text edit values.
 *
\*===========================================================================*/

void SetTextItemFloat(
	DialogPtr dlg,
	short item,
	float value)
{
	short         iType;
	Rect          iRect;
	ControlHandle iHandle;
	Str255        text;

	GetDialogItem(dlg, item, &iType, (Handle *)&iHandle, &iRect);
	sprintf(((char *)text)+1, "%5.5f", value);
	text[0] = strlen(((char *)text)+1);
	SetDialogItemText((Handle)iHandle, text);
}


/*===========================================================================*\
 *
 *	Routine:	GetTextItemFloat()
 *
 *	Comments:	Useful little subroutine for setting dialog text edit values.
 *
\*===========================================================================*/

void GetTextItemFloat(
	DialogPtr dlg,
	short item,
	float *value)
{
	short         iType;
	Rect          iRect;
	ControlHandle iHandle;
	Str255        text;

	GetDialogItem(dlg, item, &iType, (Handle *)&iHandle, &iRect);
	GetDialogItemText((Handle)iHandle, text);
	p2cstr(text);
	sscanf((char *)text, "%f", value);
}


/*===========================================================================*\
 *
 *	Routine:	UpdateStartStop()
 *
 *	Comments:	Update the text of the start stop menu item
 *
\*===========================================================================*/

static void UpdateStartStop(
	void)
{
	Str255 text;
	
	if (BirdGetIsRunning()) {
		GetIndString(text, strxMenuItems, strStop);
	}
	else {
		GetIndString(text, strxMenuItems, strStart);
	}
	SetMenuItemText(fMenuFile, fileStartStop, text);
}


/*===========================================================================*\
 *
 *	Routine:	BuildDriverMenu()
 *
 *	Comments:	Build a menu with the names of all serial drivers on the system
 *
\*===========================================================================*/

void
BuildDriverMenu(
	MenuHandle menu)
{
	short i, numDriverNames;
	Str255 driverName;
	
	if (GetNumDriverNames(&numDriverNames) != noErr) return;
	for (i = 0; i < numDriverNames; i++) {
		if (GetDriverName(i, driverName) == noErr) {
			AppendMenu(menu, driverName);
		}
		else {
			return;
		}
	}
}


/*===========================================================================*\
 *
 *	Routine:	ClickInWindow()
 *
 *	Comments:	Tracks a mouse click.
 *
\*===========================================================================*/

void ClickInWindow(
	WindowPtr wind,
	Point where)
{
	wind; /* unused */
	where; /* unused */
}


/*===========================================================================*\
 *
 *	Routine:	ErrorAlert()
 *
 *	Comments:	Puts up a error alert with an error code
 *
\*===========================================================================*/

void ErrorAlert(
	OSErr err)
{
	unsigned char numString[12];

	NumToString(err, numString);
	ParamText(numString, nil, nil, nil);
	if (err < 0) {
		/* System error (probably serial port) */
		StopAlert(alrtOSError, nil);
	}
	else if (IS_FLOCK_ERROR(err)) {
		/* Flock error */
		StopAlert(alrtBirdError, nil);
	}
	else {
		/* Bird driver error */
		StopAlert(alrtDriverError, nil);
	}
}


/*===========================================================================*\
 *
 *	Routine:	DoAppleAbout()
 *
 *	Comments:	Puts up the "About" dialog.
 *
\*===========================================================================*/

// ================= the event filter proc ==================
pascal Boolean AboutDlgFilter(
	DialogPtr theDlg,
	EventRecord *theEvent,
	short *itemHit);
pascal Boolean AboutDlgFilter(
	DialogPtr /*theDlg*/,
	EventRecord *theEvent,
	short *itemHit)
{
	if (theEvent->what == mouseDown) {
		*itemHit = 1;
		return true;
	}
	return false;
}
void DoAppleAbout(
	void)
{
	DialogPtr dlg;
	UniversalProcPtr modalFilterProc;
	short itemHit;
	Boolean done;

	dlg = GetNewDialog(dlgAbout, 0, (WindowPtr)-1);
	modalFilterProc = NewModalFilterProc(AboutDlgFilter);
	ShowWindow((WindowPtr) dlg);

	done = false;
	while (done == false) {
		ModalDialog(modalFilterProc, &itemHit);
		if (itemHit == 1) {
			done = true;
		}
	}
	DisposeDialog(dlg);
	DisposeRoutineDescriptor(modalFilterProc);
}


/*===========================================================================*\
 *
 *	Routine:	DoAppleDeskAcc()
 *
 *	Comments:	Invokes a desk accessory.
 *
\*===========================================================================*/

void DoAppleDeskAcc(
	int	item)
{
	Str255 name;
	
	GetMenuItemText(fMenuApple, item, name);
	OpenDeskAcc(name);
}


/*===========================================================================*\
 *
 *	Routine:	DoFileSerialPort()
 *
 *	Comments:	Responds to a serial port command.
 *				If a new port is set, the flock is started
 *
\*===========================================================================*/

void DoFileSerialPort(
	void)
{
	OSErr err;
	DialogRecord dlgRecord;
	DialogPtr dlg;
	UniversalProcPtr modalFilterProc;
	short itemHit;
	Boolean done, doSet;
	short iType;
	Rect iRect;
	ControlHandle iHandle;
	short port;
	int i;
	short speed;
	MenuHandle menu;

	/* Stop the flock if it is running */
	if (BirdGetIsRunning()) {
		err = BirdStopRunning();
		if (err != noErr) goto exit;
	}

	/* Create the driver pop up menu */
	menu = GetMenu(menuDriver);
	BuildDriverMenu(menu);

	/* Set up the dialog with the current values for the items */
	dlg = GetNewDialog(dlgSerialPort, &dlgRecord, (WindowPtr)-1);
	modalFilterProc = NewModalFilterProc(OkCancelDlgFilter);
	port = BirdGetPort();
	GetDialogItem(dlg, menuPort, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, port+2);
	for (i = but2400; i <= but57600; i++) {
		GetDialogItem(dlg, i, &iType, (Handle *)&iHandle, &iRect);
		SetControlValue(iHandle, 0);
	}
	switch (BirdGetPortSpeed()) {
		case baud2400:  FlipItem(dlg, but2400);  break;
		case baud4800:  FlipItem(dlg, but4800);  break;
		case baud9600:  FlipItem(dlg, but9600);  break;
		case baud19200: FlipItem(dlg, but19200); break;
		case baud57600: FlipItem(dlg, but57600); break;
	}
	
	ShowWindow((WindowPtr) dlg);

	/* Do modal dialog loop */
	done = false;
	doSet = false;
	while (done == false) {
		ModalDialog(modalFilterProc, &itemHit);
		switch (itemHit) {
			case but2400:
			case but4800:
			case but9600:
			case but19200:
			case but57600:
				for (i = but2400; i <= but57600; i++) {
					GetDialogItem(dlg, i, &iType, (Handle *)&iHandle, &iRect);
					SetControlValue(iHandle, itemHit == i);
				}
				break;
				
			case butSerialPortOk:
				/* Get the port and port speed */
				GetDialogItem(dlg, menuPort, &iType, (Handle *)&iHandle, &iRect);
				port = GetControlValue(iHandle) - 2;
				for (i = but2400; i <= but57600; i++) {
					GetDialogItem(dlg, i, &iType, (Handle *)&iHandle, &iRect);
					if (GetControlValue(iHandle)) {
						switch (i) {
							case but2400:  speed = baud2400;  break;
							case but4800:  speed = baud4800;  break;
							case but9600:  speed = baud9600;  break;
							case but19200: speed = baud19200; break;
							case but57600: speed = baud57600; break;
						}
						break;
					}
				}
				doSet = true;
				done = true;
				break;
				
			case butSerialPortCancel:
				done = true;
				break;
		}
	}
	
	/* Remove the dialog */
	DisposeDialog(dlg);
	DisposeRoutineDescriptor(modalFilterProc);

	/* Activate the new port */
	if (doSet) { 
		err = BirdSetPort(port, speed);
		if (err != noErr) { goto exit; }
		if (BirdGetPort() == -1) { goto exit; }
		/* Set settings from preferences */
		err = BirdSetConfiguration(&gPreferences.configuration);
		if (err != noErr) { goto exit; }
	}
	/* Start running the flock */
	err = BirdStartRunning();	
	if (err != noErr) { goto exit; }

exit:
	if (err != noErr) {
		ErrorAlert(err);
	}
}


/*===========================================================================*\
 *
 *	Routine:	DoFlockConfig()
 *
 *	Comments:	Responds to a configuration command.
 *				If a new configuration is set, the flock is started
 *
\*===========================================================================*/

// ================= the event filter proc ==================
pascal Boolean FlockConfigDlgFilter(
	DialogPtr theDlg,
	EventRecord *theEvent,
	short *itemHit);
pascal Boolean FlockConfigDlgFilter(
	DialogPtr theDlg,
	EventRecord *theEvent,
	short *itemHit)
{
	short iType;
	Rect iRect;
	Handle iHandle;
	Boolean displayValues;
	float voltage, scanRate;

	GetDialogItem(theDlg, checkSyncToMonitor, &iType, &iHandle, &iRect);
	displayValues = GetControlValue((ControlHandle)iHandle);
	if (displayValues) {
		BirdGetSyncData(&voltage, &scanRate);
		SetTextItemFloat(theDlg, statVoltage, voltage);
		SetTextItemFloat(theDlg, statScanRate, scanRate);
	}
	else {
		GetDialogItem(theDlg, statVoltage, &iType, &iHandle, &iRect);
		SetDialogItemText((Handle)iHandle, "\p");
		GetDialogItem(theDlg, statScanRate, &iType, &iHandle, &iRect);
		SetDialogItemText((Handle)iHandle, "\p");
	}

	if (theEvent->what == keyDown) {
		switch ((theEvent->message) & charCodeMask) {
			case 0x0D:
			case 0x03:
				*itemHit = 1;
				return true;
			case 0x1B:
				*itemHit = 2;
				return true;
		}
	}
	return false;
}
void DoFlockConfig(
	void)
{
	OSErr err;
	DialogPtr dlg;
	UniversalProcPtr modalFilterProc;
	short iType;
	Rect iRect;
	ControlHandle iHandle;
	Str15 text;
	short itemHit;
	Boolean done, doSet;
	TBirdInformation info;
	TBirdConfiguration config;
	int value;
	float voltage, scanRate;
	Boolean wasRunning;

	wasRunning = BirdGetIsRunning();
	/* Stop running the flock and get current configuration */
	err = BirdGetConfiguration(&config);
	if (err != noErr) goto exit;
	err = BirdGetInformation(&info);
	if (err != noErr) goto exit;

	/* Set up the dialog with the current values for the items */
	dlg = GetNewDialog(dlgFlockConfig, 0, (WindowPtr)-1);
	modalFilterProc = NewModalFilterProc(FlockConfigDlgFilter);
	/* Filter on/off status */
	GetDialogItem(dlg, radioACNarrow, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, config.ACNarrowOn);
	GetDialogItem(dlg, radioACWide, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, config.ACWideOn);
	GetDialogItem(dlg, checkDC, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, config.DCOn);
	/* Measurement rate */
	SetTextItemFloat(dlg, editMeasurementRate, ((float)config.measurementRate) / ((float)256));
	/* Sudden output elimination */
	GetDialogItem(dlg, checkSuddenOutputElimination, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, config.suddenOutputElimination);
	/* Sync type */
	GetDialogItem(dlg, checkSyncToMonitor, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, (config.sync == sync_CRT1) || (config.sync == sync_CRT2));
	GetDialogItem(dlg, statVoltage, &iType, (Handle *)&iHandle, &iRect);
	SetDialogItemText((Handle)iHandle, "\p");
	GetDialogItem(dlg, statScanRate, &iType, (Handle *)&iHandle, &iRect);	
	SetDialogItemText((Handle)iHandle, "\p");
	/* Range */
	GetDialogItem(dlg, radioRange36, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, config.range == range_36);
	GetDialogItem(dlg, radioRange72, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, config.range == range_72);
	
	/* Number of birds */
	SetTextItemInt(dlg, statNumBirdsOperating, info.numBirdsOperating);
	SetTextItemInt(dlg, statNumBirdsFound, info.numBirdsFound);
	/* Flock software revision */
	GetDialogItem(dlg, statSoftwareRev, &iType, (Handle *)&iHandle, &iRect);
	sprintf(((char *)text)+1,
	        "%d.%d",
	        info.softwareRevInt,
	        info.softwareRevFrac);
	text[0] = strlen(((char *)text)+1);
	SetDialogItemText((Handle)iHandle, text);
	/* Crystal speed */
	SetTextItemInt(dlg, statCrystalSpeed, info.crystalSpeed);

	ShowWindow((WindowPtr) dlg);

	/* Do modal dialog loop */
	done = false;
	doSet = false;
	while (done == false) {
		ModalDialog(modalFilterProc, &itemHit);
		switch (itemHit) {
			case radioACNarrow:
			case radioACWide:
				GetDialogItem(dlg, radioACNarrow, &iType, (Handle *)&iHandle, &iRect);
				SetControlValue(iHandle, 0);
				GetDialogItem(dlg, radioACWide, &iType, (Handle *)&iHandle, &iRect);
				SetControlValue(iHandle, 0);
				GetDialogItem(dlg, itemHit, &iType, (Handle *)&iHandle, &iRect);
				SetControlValue(iHandle, 1);
				break;

			case checkDC:
			case checkSuddenOutputElimination:
				FlipItem(dlg, itemHit);
				break;

			case checkSyncToMonitor:
				FlipItem(dlg, itemHit);
				GetDialogItem(dlg, checkSyncToMonitor, &iType, (Handle *)&iHandle, &iRect);
				if (!GetControlValue(iHandle)) {
					GetDialogItem(dlg, statVoltage, &iType, (Handle *)&iHandle, &iRect);
					SetDialogItemText((Handle)iHandle, "\p");
					GetDialogItem(dlg, statScanRate, &iType, (Handle *)&iHandle, &iRect);	
					SetDialogItemText((Handle)iHandle, "\p");
				}
				break;

			case radioRange36:
			case radioRange72:
				GetDialogItem(dlg, radioRange36, &iType, (Handle *)&iHandle, &iRect);
				SetControlValue(iHandle, 0);
				GetDialogItem(dlg, radioRange72, &iType, (Handle *)&iHandle, &iRect);
				SetControlValue(iHandle, 0);
				GetDialogItem(dlg, itemHit, &iType, (Handle *)&iHandle, &iRect);
				SetControlValue(iHandle, 1);
				break;

			case butFlockConfigOk:
				/* Filter on/off status */
				GetDialogItem(dlg, radioACNarrow, &iType, (Handle *)&iHandle, &iRect);
				config.ACNarrowOn = GetControlValue(iHandle);
				GetDialogItem(dlg, radioACWide, &iType, (Handle *)&iHandle, &iRect);
				config.ACWideOn = GetControlValue(iHandle);
				GetDialogItem(dlg, checkDC, &iType, (Handle *)&iHandle, &iRect);
				config.DCOn = GetControlValue(iHandle);
				/* Measurement rate */
				GetTextItemInt(dlg, editMeasurementRate, &value);
				config.measurementRate = value * 256;
				/* Sudden output elimination */
				GetDialogItem(dlg, checkSuddenOutputElimination, &iType, (Handle *)&iHandle, &iRect);
				config.suddenOutputElimination = GetControlValue(iHandle);
				/* Sync setting */
				config.sync = sync_None;
				GetDialogItem(dlg, checkSyncToMonitor, &iType, (Handle *)&iHandle, &iRect);
				if (GetControlValue(iHandle)) {
					/* Read the voltage and scan rate values */
					BirdGetSyncData(&voltage, &scanRate);
					if (voltage > 1) {
						if ((scanRate > 72) && (scanRate < 144)) { 
							config.sync = sync_CRT1;
						}
						else if ((scanRate > 50) && (scanRate <= 72)) {
							config.sync = sync_CRT2;
						}
					}
				}
				/* Range */
				GetDialogItem(dlg, radioRange36, &iType, (Handle *)&iHandle, &iRect);
				config.range = GetControlValue(iHandle) ? range_36 : range_72;
				
				/* Activate the settings */
				done = true;
				doSet= true;
				break;
				
			case butFlockConfigCancel:
				done = true;
				break;
		}
	}
	/* Clean up dialog */
	DisposeDialog(dlg);
	DisposeRoutineDescriptor(modalFilterProc);

	err = BirdSetConfiguration(&config);
	if (err != noErr) goto exit;
	if (wasRunning) {
		err = BirdStartRunning();
		if (err != noErr) goto exit;
	}

exit:
	if (err != noErr) {
		ErrorAlert(err);
	}
}


/*===========================================================================*\
 *
 *	Routine:	DoQD3DConfig()
 *
 *	Comments:	Responds to a configuration command.
 *
\*===========================================================================*/

void DoQD3DConfig(
	void)
{
	OSErr err;
	DialogPtr dlg;
	UniversalProcPtr modalFilterProc;
	short iType;
	Rect iRect;
	ControlHandle iHandle;
	short itemHit;
	Boolean done, doSet;
	float inch2qd3dX, inch2qd3dY, inch2qd3dZ;
//	TBirdOperation operation;
	TBirdMode mode;
	Boolean controlCursor;
	Boolean wasRunning;

	/* Stop running the flock */
	wasRunning = BirdGetIsRunning();
	err = BirdStopRunning();
	if (err != noErr) goto exit;

	/* Set up the dialog with the current values for the items */
	dlg = GetNewDialog(dlgQD3DConfig, 0, (WindowPtr)-1);
	modalFilterProc = NewModalFilterProc(OkCancelDlgFilter);
	BirdGetInch2QD3D(&inch2qd3dX, &inch2qd3dY, &inch2qd3dZ);
	SetTextItemFloat(dlg, editInch2qd3dX, inch2qd3dX);
	SetTextItemFloat(dlg, editInch2qd3dY, inch2qd3dY);
	SetTextItemFloat(dlg, editInch2qd3dZ, inch2qd3dZ);
//	operation = BirdGetOperation();
//	GetDialogItem(dlg, radioPoll, &iType, (Handle *)&iHandle, &iRect);
//	SetControlValue(iHandle, mode == mode_Poll);
//	GetDialogItem(dlg, radioStream, &iType, (Handle *)&iHandle, &iRect);
//	SetControlValue(iHandle, mode == mode_Stream);
	mode = BirdGetMode();
	GetDialogItem(dlg, radioRelative, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, mode == mode_Relative);
	GetDialogItem(dlg, radioAbsolute, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, mode == mode_Absolute);
	controlCursor = BirdGetControlCursor();
	GetDialogItem(dlg, checkControlCursor, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, controlCursor);
	ShowWindow((WindowPtr) dlg);

	/* Do modal dialog loop */
	done = false;
	doSet = false;
	while (!done) {
		ModalDialog(modalFilterProc, &itemHit);
		switch (itemHit) {
//			case radioPoll:
//			case radioStream:
//				GetDialogItem(dlg, radioPoll, &iType, (Handle *)&iHandle, &iRect);
//				SetControlValue(iHandle, 0);
//				GetDialogItem(dlg, radioStream, &iType, (Handle *)&iHandle, &iRect);
//				SetControlValue(iHandle, 0);
//				GetDialogItem(dlg, itemHit, &iType, (Handle *)&iHandle, &iRect);
//				SetControlValue(iHandle, 1);
//				break;
				
			case radioRelative:
			case radioAbsolute:
				GetDialogItem(dlg, radioRelative, &iType, (Handle *)&iHandle, &iRect);
				SetControlValue(iHandle, 0);
				GetDialogItem(dlg, radioAbsolute, &iType, (Handle *)&iHandle, &iRect);
				SetControlValue(iHandle, 0);
				GetDialogItem(dlg, itemHit, &iType, (Handle *)&iHandle, &iRect);
				SetControlValue(iHandle, 1);
				break;
				
			case checkControlCursor:
				FlipItem(dlg, itemHit);
				break;
				
			case butQD3DConfigOk:
				GetTextItemFloat(dlg, editInch2qd3dX, &inch2qd3dX);
				GetTextItemFloat(dlg, editInch2qd3dY, &inch2qd3dY);
				GetTextItemFloat(dlg, editInch2qd3dZ, &inch2qd3dZ);
//				GetDialogItem(dlg, radioPoll, &iType, (Handle *)&iHandle, &iRect);
//				operation = GetControlValue(iHandle) ? mode_Poll : mode_Stream;
				GetDialogItem(dlg, radioRelative, &iType, (Handle *)&iHandle, &iRect);
				mode = GetControlValue(iHandle) ? mode_Relative : mode_Absolute;
				GetDialogItem(dlg, checkControlCursor, &iType, (Handle *)&iHandle, &iRect);
				controlCursor = GetControlValue(iHandle);
				doSet = true;
				done = true;
				break;
				
			case butQD3DConfigCancel:
				done = true;
				break;
		}
	}	
	/* Clean up */
	DisposeDialog(dlg);
	DisposeRoutineDescriptor(modalFilterProc);

	/* Activate new settings */
	if (doSet) {
		BirdSetInch2QD3D(inch2qd3dX, inch2qd3dY, inch2qd3dZ);
//		err = BirdSetOperation(operation);
//		if (err != noErr) goto exit;
		BirdSetMode(mode);
		BirdSetControlCursor(controlCursor);
	}
	/* Start the flock (if it was running) */
	if (wasRunning) {
		err = BirdStartRunning();
		if (err != noErr) goto exit;
	}

exit:
	if (err != noErr) {
		ErrorAlert(err);
	}
}


/*===========================================================================*\
 *
 *	Routine:	DoFileQuit()
 *
 *	Comments:	Responds to a Quit command.
 *
\*===========================================================================*/

void DoFileQuit(
	void)
{
	fExit = true;
}


/*===========================================================================*\
 *
 *	Routine:	DoMenuCommand()
 *
 *	Comments:	Switches out to handle a menu selection.
 *
\*===========================================================================*/

void DoMenuCommand(
	long itemVal)
{
	int item;
	OSErr err;
	
	item = LoWord(itemVal);
	
	switch (HiWord(itemVal)) {
		case menuApple:
			switch(item) {
				case appleAbout: DoAppleAbout();       break;
				default:         DoAppleDeskAcc(item); break;
			}
			break;
		
		case menuFile:
			switch(item) {
				case fileSerialPort:
					DoFileSerialPort();
					break;
				case fileFlockConfig:
					DoFlockConfig();
					break;
				case fileQD3DConfig:
					DoQD3DConfig();
					break;
				case fileStartStop:
					if (BirdGetIsRunning()) {
						err = BirdStopRunning();
					}
					else {
						err = BirdStartRunning();
					}
					if (err) ErrorAlert(err);
					UpdateStartStop();
					break;
				case fileQuit:
					DoFileQuit();
					break;
			}
			break;
	}
	
	HiliteMenu(0);
}


/*===========================================================================*\
 *
 *	Routine:	WindowDrag()
 *
 *	Comments:	Drags the window.
 *
\*===========================================================================*/

void WindowDrag(
	WindowPtr wind,
	Point where)
{
	DragWindow(wind, where, &qd.screenBits.bounds);
}


/*===========================================================================*\
 *
 *	Routine:	DoEventMouseDown()
 *
 *	Comments:	Responds to a mouse-down event in menu bar or system window.
 *
\*===========================================================================*/

void DoEventMouseDown(
	void)
{
	WindowPtr wind;
	
	switch(FindWindow(fEvent.where, &wind)) {
		case inMenuBar:
			UpdateStartStop();
			DoMenuCommand(MenuSelect(fEvent.where));
			break;
		case inSysWindow:
			SystemClick(&fEvent, wind);
			break;
		case inContent:
			ClickInWindow(wind, fEvent.where);
			break;
		case inDrag:
			WindowDrag(wind, fEvent.where);
			break;
	}
}


/*===========================================================================*\
 *
 *	Routine:	DoEventKeyDown()
 *
 *	Comments:	Processes a key down by checking for a menu key.
 *
\*===========================================================================*/

void DoEventKeyDown(
	void)
{
	char ch;
	
	ch = fEvent.message & charCodeMask;
	
	if (fEvent.modifiers & cmdKey)
		DoMenuCommand(MenuKey(ch));
}


/*===========================================================================*\
 *
 *	Routine:	DoEventUpdate()
 *
 *	Comments:	Processes an update event.
 *
\*===========================================================================*/

void DoEventUpdate(
	WindowPtr wind)
{
	BeginUpdate(wind);
	EndUpdate(wind);
}


/*===========================================================================*\
 *
 *	Routine:	DoEventActivate()
 *
 *	Comments:	Processes an activate event.
 *
\*===========================================================================*/

void DoEventActivate(
	WindowPtr wind)
{
	wind; /* unused */
}


/*===========================================================================*\
 *
 *	Routine:	Init()
 *
 *	Comments:	Sets up menus, etc.
 *
\*===========================================================================*/

void Init(
	void)
{
	OSErr err;
	TQ3Status status;
	
	fExit = false;

	MaxApplZone();
	
	InitGraf(&qd.thePort);
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(nil);
	InitCursor();
	
	GetNextEvent(everyEvent, &fEvent);
	GetNextEvent(everyEvent, &fEvent);
	GetNextEvent(everyEvent, &fEvent);
	
	/* Form the menubar */
	InsertMenu(fMenuApple = GetMenu(menuApple), 0);
	InsertMenu(fMenuFile = GetMenu(menuFile), 0);
	AppendResMenu(fMenuApple, 'DRVR');
	DrawMenuBar();
		
	SetCursor(&qd.arrow);
	
	/* Create the serial driver name table */
	NewNameTable();
	
	/* Load the preferences of the driver (if option key down defaults) */
	GetPreferencesFile(Creator_Preferences, "\pBird Preferences", &gPrefFile);
	if (fEvent.modifiers & optionKey) {
		SysBeep(0);
		GetDefaultPreferences(&gPreferences);
	}
	else {
		LoadPreferences(&gPrefFile, &gPreferences);
	}
	
	/* Initialize Escher */
	status = Q3Initialize();
	if (status != kQ3Success) { err = err_QD3DInitialise; goto exit; }

	/* Set the error, warning and notice methods */
	status = Q3Error_Register(ErrorMethod, 0);
	if (status != kQ3Success) { err = err_QD3DInitialise; goto exit; }
	status = Q3Warning_Register(WarningMethod, 0);
	if (status != kQ3Success) { err = err_QD3DInitialise; goto exit; }
	status = Q3Notice_Register(NoticeMethod, 0);
	if (status != kQ3Success) { err = err_QD3DInitialise; goto exit; }
	
	/* Initialize the device with the settings from the preferences */
	err = BirdInit(gPreferences.port,
	               gPreferences.speed,
	               &gPreferences.configuration,
#if OPERATION_MODE
	               mode_Poll,
#endif
	               gPreferences.mode,
	               gPreferences.inch2qd3dX,
	               gPreferences.inch2qd3dY,
	               gPreferences.inch2qd3dZ,
	               gPreferences.controlCursor);
	if (err != noErr) { goto exit; }

exit:
	if (err != noErr) {
		ErrorAlert(err);
	}
}


/*===========================================================================*\
 *
 *	Routine:	Loop()
 *
 *	Comments:	Gets and processes events.
 *
\*===========================================================================*/

void Loop(	
	void)
{
	long sleep;
#if OPERATION_MODE
	OSErr err;
#endif
	
	sleep = 0;
	do {
		WaitNextEvent(everyEvent, &fEvent, sleep, nil);
		switch(fEvent.what) {
			case mouseDown:
				DoEventMouseDown();
				break;
			case keyDown:
				DoEventKeyDown();
				break;
			case activateEvt:
				DoEventActivate((WindowPtr) fEvent.message);
				break;
			case updateEvt:
				DoEventUpdate((WindowPtr) fEvent.message);
				break;
		}
		sleep = 60;
#if OPERATION_MODE
		if (BirdGetOperation() == mode_Poll) {
			err = BirdPoll();
			sleep = 1;
		}
#endif
		if (fLastError != kQ3WarningNone) {
			ShowError();
			fFirstError = kQ3ErrorNone;
			fLastError = kQ3ErrorNone;
		}
		if (fLastWarning != kQ3WarningNone) {
			ShowWarning();
			fFirstWarning = kQ3WarningNone;
			fLastWarning = kQ3WarningNone;
		}
		if (fLastNotice != kQ3NoticeNone) {
			ShowNotice();
			fFirstNotice = kQ3NoticeNone;
			fLastNotice	= kQ3NoticeNone;
		}
	} while (!fExit);
}


/*===========================================================================*\
 *
 *	Routine:	Term()
 *
 *	Comments:	Cleans up before exit.
 *
\*===========================================================================*/

void Term(
	void)
{
	gPreferences.port = BirdGetPort();
	gPreferences.speed = BirdGetPortSpeed();
	BirdGetConfiguration(&gPreferences.configuration);
//	gPreferences.operation = BirdGetOperation();
	gPreferences.mode = BirdGetMode();
	BirdGetInch2QD3D(&gPreferences.inch2qd3dX,
	                 &gPreferences.inch2qd3dY,
	                 &gPreferences.inch2qd3dZ);
	gPreferences.controlCursor = BirdGetControlCursor();
	SavePreferences(&gPrefFile, &gPreferences);
	BirdTerm();
	
	Q3Exit();
	
	/* Dispose the serial driver name table */
	DisposeNameTable();
}

void ErrorMethod(
	TQ3Error firstError,
	TQ3Error lastError,
	long /*reference*/)
{
	fFirstError = firstError;
	fLastError = lastError;
	fNumErrors++;
}
void WarningMethod(
	TQ3Warning firstWarning,
	TQ3Warning lastWarning,
	long /*reference*/)
{
	fFirstWarning = firstWarning;
	fLastWarning = lastWarning;
	fNumWarnings++;
}
void NoticeMethod(
	TQ3Notice firstNotice,
	TQ3Notice lastNotice,
	long /*reference*/)
{
	fFirstNotice = firstNotice;
	fLastNotice = lastNotice;
	fNumNotices++;
}


void ShowError()
{
	short answer;
	Str15 first, last, num;
	
	NumToString(fFirstError, first);
	NumToString(fLastError, last);
	NumToString(fNumErrors, num);
	ParamText(first, last, num, NULL);
	answer = StopAlert(alrtQD3DError, NULL);
	if (answer == 2) {
		BirdTerm();
		Q3Exit();
		ExitToShell();
	}
}
void ShowWarning()
{
	short answer;
	Str15 first, last, num;
	
	NumToString(fFirstWarning, first);
	NumToString(fLastWarning, last);
	NumToString(fNumWarnings, num);
	ParamText(first, last, num, NULL);
	answer = CautionAlert(alrtQD3DWarning, NULL);
	if (answer == 2) {
		BirdTerm();
		Q3Exit();
		ExitToShell();
	}
}
void ShowNotice()
{
	short answer;
	Str15 first, last, num;
	
	NumToString(fFirstNotice, first);
	NumToString(fLastNotice, last);
	NumToString(fNumNotices, num);
	ParamText(first, last, num, NULL);
	answer = NoteAlert(alrtQD3DNotice, NULL);
	if (answer == 2) {
		BirdTerm();
		Q3Exit();
		ExitToShell();
	}
}

/*===========================================================================*\
 *
 *	Routine:	main()
 *
 *	Comments:	xxx
 *
\*===========================================================================*/

void main(
	void)
{
	Init();
//	ErrorMethod(kQ3ErrorNone, kQ3ErrorNotInitialized, 0);
//	WarningMethod(kQ3WarningNone, kQ3WarningNoObjectSupportForDrawMethod, 0);
//	NoticeMethod(kQ3NoticeNone, kQ3NoticeDataAlreadyEmpty, 0);
	Loop();
	Term();
}