/******************************************************************************
 **																			 **
 ** 	Module:		DriverDynasight.c				    					 **
 ** 																		 **
 ** 																		 **
 ** 	Purpose: 	Connects to the Dynasight device						 **
 **					Created by Maarten Gribnau								 **
 **					email: m.w.gribnau@io.tudelft.nl						 **
 **					Loosely based on code for Magellan driver.				 **
 **					Was part of Quickdraw 3D SDK example code.				 **
 **					The original notice is included below.					 **
 ** 																		 **
 ** 																		 **
 ** 																		 **
 ** 																		 **
 *****************************************************************************/

/** 				Copyright (C) 1992-96 Apple Computer, Inc.				 **
 **					All rights reserved.									 **/

/* this code is supplied on the understanding that it is sample code
 * and is completly unfit for any particular purpose.  In addition, this
 * sample is unsupported.
 *
 * Please send questions to DEVSUPPORT, send bugs to NICKT
 *
 * This project builds an app which is the "driver" app.  You need to 
 * run this before any application that uses this driver.  This app will 
 * sit there and poll the input device for changes.  When it receives changes
 * in orientation it will update the controller.  The app that uses the controller
 * (a sample is supplied with this driver, the sample is called tryTracker) must
 * poll the controller to get the orientation changes.
 *
 * Please contact stephen ilberg at Logitech for more information on the Magellan product
 */
 
 
#include "DriverDynasight.h"

#include <Devices.h>
#include <Serial.h>
#include <Errors.h>
#include <QD3D.h>
#include <QD3DController.h>

#include "SerialDriverArbitration.h"
#include "SerialDriverNames.h"



/******************************************************************************
 **																			 **
 **							Variables and Forward Declarations				 **
 **																			 **
 *****************************************************************************/


#define BUFFER_LENGTH		256
#define PACKET_LENGTH		8
/* Target not in track, search in progress */
#define SS_SEARCH			0x00
/* Target lost, attempting to reaquire */
#define SS_COAST			0x01
/* Track in progress, conditions are marginal, eminent loss-of-track possible */
#define SS_CAUTION			0x02	
/* Track in progress, signal-to-noise ratio satisfactory */
#define SS_TRACK			0x03	
/* The SYNC input/output line is high or undriven (50 msec debounce delay) 
 * In StereoSync modes: Internal synchronization
 */
#define SYNC_IO_0			0x00		
/* The SYNC input/output line is driven low (50 msec debounce delay) 
 * In StereoSync modes: External synchronization
 */
#define SYNC_IO_1			0x01		

typedef struct {
	SInt16	x;
	SInt16	y;
	SInt16	z;
} TDynaPoint;


static TQ3ControllerRef		fControllerRef = NULL;
static TDynaConfiguration	fConfiguration =
{ -1, oper_Poll, 0, 0, 0, 1, 1, 1, false };
static short				fSerInp = 0;
static short				fSerOut = 0;
static Boolean				fIsRunning = false;
static DeferredTask			fDefTask;
static IOParam				fSerParm;
static Byte					fBuffer[BUFFER_LENGTH];
static long					fBufferCount = 0;

static Byte				fSyncIO = SYNC_IO_0;
static Byte				fSyncStatus = SS_SEARCH;
static TDynaPoint		fPos;	


/******************************************************************************
 **									Prototypes
 *****************************************************************************/

static OSErr SerRecvFlush(
	short serPort);

static pascal void ASyncCompletion(
	ParmBlkPtr paramBlock);

static pascal void DeferredDecode(
	long dtParam);

static void AccumulateBuffer(
	Byte *buffer,
	long *count);

static void DecodePackage(
	Byte data[8],
	TDynaPoint *point);

static void UpdateController(
	void);

static OSErr SerialPortInit(
	short port,
	short *serPortIn,
	short *serPortOut);

static OSErr SerialPortTerm(
	short serPortIn,
	short serPortOut);

static OSErr SerialPortSet(
	short port);



/******************************************************************************
 **									Private  Routines
 *****************************************************************************/

/*===========================================================================*\
 *
 *	Routine:	SerRecvFlush()
 *
 *	Comments:	Clear the serial input port buffer 
 *
\*===========================================================================*/

static OSErr SerRecvFlush(
	short serPort)
{
	OSErr		err = noErr;
	long		count;
	char		buffer[BUFFER_LENGTH];
	Boolean		ready = false;
	
	while (!ready) {
		err = SerGetBuf(serPort, &count);
		if (err != noErr) break;
		if (count > 0) {
			if (count > BUFFER_LENGTH) {
				count = BUFFER_LENGTH;
			}
			err = FSRead(serPort, &count, (Ptr) buffer);
			if (err != noErr) break;
		}
		else {
			ready = true;
		}
	}
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	ASyncCompletion()
 *
 *	Comments:	This is the serial input async completion routine. It adjusts
 *				the buffer size with the amount of incoming data and invokes a
 *				deferred task to parse it.
 *
\*===========================================================================*/

static pascal void ASyncCompletion(
	ParmBlkPtr /*paramBlock*/)
{
	/* Bail out on KillIO */
	if (fSerParm.ioResult == abortErr) {
		return;
	}
	
	/* We have a few more characters */
	fBufferCount += fSerParm.ioActCount;

	/* Bag the buffer on serial error */	
	if (fSerParm.ioResult != noErr) {
		fBufferCount = 0;
	}
	
	/* Install the deferred task */
	DTInstall(&fDefTask);
}


/*===========================================================================*\
 *
 *	Routine:	DeferredDecode()
 *
 *	Comments:	This is the deferred task that is invoked by the serial input
 *				async completion routine.  It tries to parse the data in the
 *				serial input buffer, and issues	another read for the next
 *				block of data.
 *
\*===========================================================================*/

pascal void DeferredDecode(
	long /*dtParam*/)
{
	long reqCount;
	TDynaPoint oldPos = fPos;		
	
	AccumulateBuffer(fBuffer, &fBufferCount);
	if ((fPos.x != oldPos.x) || (fPos.y != oldPos.y) || (fPos.z != oldPos.z)) {
		UpdateController();
	}

	reqCount = PACKET_LENGTH - fBufferCount;
	if ((reqCount > 0) && (fBufferCount < (BUFFER_LENGTH - reqCount))) {
		fSerParm.ioBuffer	= ((Ptr) fBuffer) + fBufferCount;
		fSerParm.ioReqCount	= reqCount;
	}
	else {
		/* Error situation, reset the buffer */
		fSerParm.ioBuffer	= (Ptr) fBuffer;
		fSerParm.ioReqCount	= PACKET_LENGTH;
		fBufferCount = 0;
	}
	PBReadAsync((ParmBlkPtr) &fSerParm);

}


/*===========================================================================*\
 *
 *	Routine:	AccumulateBuffer()
 *
 *	Comments:	All the packages in the buffer are processed and the controller
 *				is updated
 *
\*===========================================================================*/

static void AccumulateBuffer(
	Byte *buffer,
	long *count)
{
	Byte *buf = buffer;
//	long i;
//	Boolean isMarker;
	/* TEMP */
	long counter = 0;

	while (*count > 0) {
		/* TEMP */
		if (++counter > 100) {
			DebugStr("\pLooped more than 100 times");
		}

		if ((*buf & 0xF0) == 0x80) {
			/* buf points to byte with the sync pattern 1000XXXX */
			if (*count < PACKET_LENGTH) {
				/* Not enough bytes in buffer, break out of loop */
				break;
			}
			else {
				/* Enough bytes are in, check if this is a valid package */
				if (((buf[1] & 0xF0) == 0x80) && /* 2nd byte hass sync pattern too */
				    ((buf[0] & 0x0C) == 0x00)) { /* message type field is zero */
				    DecodePackage(buf, &fPos);
					*count -= PACKET_LENGTH;
					buf += PACKET_LENGTH;
				}			
			}
		}
		else {
			/* Wait for sync byte */
			(*count)--;
			buf++;
		}
	}
	
	if (*count > 0) {
		/* Move the bytes not processed to the start of the buffer */
		BlockMove(buf, buffer, *count);
	}
	else {
		/* All bytes in the buffer are processed */
		*count = 0;
	}	
}

	
/*===========================================================================*\
 *
 *	Routine:	DecodePackage()
 *
 *	Comments:	Get the current position from the data bytes
 *				Update the current sync io and sync status
 *
\*===========================================================================*/

static void DecodePackage(
	Byte data[8],
	TDynaPoint *point)
{
	Byte exp = data[0] & 0x03;
	
	fSyncIO = (data[1] & 0x08) >> 3;
	fSyncStatus = (data[1] & 0x03);
	
	point->x = (data[2] << 8) | data[3];
	point->y = (data[4] << 8) | data[5];
	point->z = (data[6] << 8) | data[7];
	point->x <<= exp;
	point->y <<= exp;
	point->z <<= exp;
}


/*===========================================================================*\
 *
 *	Routine:	UpdateController()
 *
 *	Comments:	Update the QD3D controller position
 *
\*===========================================================================*/

static void UpdateController()
{
	TQ3Boolean track2DCursor;
	TQ3Point3D pos;
	
	Q3Controller_Track2DCursor(fControllerRef, &track2DCursor);
	if ((track2DCursor == kQ3False) ||
	    ((track2DCursor == kQ3True) && fConfiguration.controlCursor)) {
		pos.x =  (fConfiguration.scaleX * ((float)fPos.x) / (20*25.4)) + fConfiguration.offsetX / 2.54;
		pos.y =  (fConfiguration.scaleY * ((float)fPos.y) / (20*25.4)) + fConfiguration.offsetY / 2.54;
		pos.z =  (fConfiguration.scaleZ * ((float)fPos.z) / (20*25.4)) + fConfiguration.offsetZ / 2.54;
		printf("%f %f %f\n", pos.x, pos.y, pos.z);
		Q3Controller_SetTrackerPosition(fControllerRef, &pos);
	}
}


/*===========================================================================*\
 *
 *	Routine:	SerialPortInit()
 *
 *	Comments:	Initialize the serial port for communication with the device
 *
\*===========================================================================*/

static OSErr SerialPortInit (
	short port,
	short *serPortIn,
	short *serPortOut)
{
	SerShk flags;
	OSErr err;
	Str255 name;

	*serPortIn = 0;
	*serPortOut = 0;
	
	if (port < 0) { err = paramErr; goto bail; }
	err = GetOutputDriverName(port, name);
	if (err != noErr) goto bail;
	if (DriverIsOpen(name)) { err = openErr; goto bail; }
	err = OpenDriver(name, serPortOut);
	if (err != noErr) goto bail;
	err = GetInputDriverName(port, name);
	if (err != noErr) goto bail;
	if (DriverIsOpen(name)) { err = openErr; goto bail; }
	err = OpenDriver(name, serPortIn);
	if (err != noErr) goto bail;
	
	/* Setup ports communications parameters */
	flags.fXOn = false;		/* output XON/XOFF setting */
	flags.fCTS = false;		/* hardware CTS setting */
	flags.xOn  = 0x11;		/* dc1  ^Q */
	flags.xOff = 0x13;		/* dc3  ^S */
	flags.errs = 0;			/* post no error events */
	flags.evts = 0;			/* post no error events */
	flags.fInX = false;		/* input XON/XOFF setting */
	flags.fDTR = false;		/* hardware DTR setting */	
	err = SerHShake(*serPortIn, &flags);
	if (err != noErr) goto bail;
	
	/* 19200 Baud 1 Stopbit No Parity 8 Databits */
	err = SerReset(*serPortIn, baud19200|data8|stop10|noParity);
	if (err != noErr) goto bail;
	
	return err;

bail:
	if (*serPortIn != 0) {
		KillIO(*serPortIn);
		CloseDriver(*serPortIn);
		*serPortIn = 0;
	}
	if (*serPortOut != 0) {
		KillIO(*serPortOut);
		CloseDriver(*serPortOut);
		*serPortOut = 0;
	}
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	SerialPortTerm()
 *
 *	Comments:	Close the serial port.
 *
\*===========================================================================*/

static OSErr SerialPortTerm (
	short serPortIn,
	short serPortOut)
{
	OSErr err, rval = noErr;
	
	if (serPortIn != 0) {
		err = KillIO(serPortIn);
		if (err) rval = err;
		err = CloseDriver(serPortIn);
		if (err) rval = err;
	}
	if (serPortOut != 0) {
		err = KillIO(serPortOut);
		if (err) rval = err;
		err = CloseDriver(serPortOut);
		if (err) rval = err;
	}

	return rval;
}


/*===========================================================================*\
 *
 *	Routine:	SerialPortSet()
 *
 *	Comments:	Set the current serial port
 *
\*===========================================================================*/

static OSErr SerialPortSet(
	short port)
{
	OSErr err = noErr;
	
	/* Close the current port */
	if (fConfiguration.serialPort != -1) {
		err = SerialPortTerm(fSerInp, fSerOut);
		fSerInp = 0;
		fSerOut = 0;
		fConfiguration.serialPort = -1;
		if (err != noErr) {	goto exit; }
	}

	/* Open the new port */
	if (port != -1) {
		err = SerialPortInit(port, &fSerInp, &fSerOut);
		if (err != noErr) { goto exit; }
		fConfiguration.serialPort = port;
		if (err != noErr) { goto exit; }
	}
	
exit:
	return err;
}


/******************************************************************************
 **									Public Routines
 *****************************************************************************/

/*===========================================================================*\
 *
 *	Routine:	DynaInit()
 *
 *	Comments:	Create the QD3D controller, set the configuration
 *
\*===========================================================================*/

OSErr DynaInit(
	TDynaConfiguration *config)
{
	OSErr err;
	TQ3ControllerData controllerData;
	
	controllerData.signature		= "Dynasight Controller, Maarten Gribnau 1997";
	controllerData.valueCount		= 0;
	controllerData.channelCount		= 0;
	controllerData.channelGetMethod	= NULL ;
	controllerData.channelSetMethod	= NULL ;
	fControllerRef = Q3Controller_New(&controllerData);
	err = err_Controller;
	if (fControllerRef == NULL) goto exit;
	
	err = DynaSetConfiguration(config);

exit:
	return err;
}


#if POLL_MODE
/*===========================================================================*\
 *
 *	Routine:	DynaPoll()
 *
 *	Comments:	If in poll mode and running, will poll the serial port for new
 *				data
 *
\*===========================================================================*/

OSErr DynaPoll(
	void)
{
	OSErr err = noErr;
	long available, count;
	TDynaPoint oldPos;
	
	if ((fConfiguration.operationMode != oper_Poll) || (!fIsRunning)) goto exit;

	/* Get the number of bytes waiting to be processed */
	err = SerGetBuf(fSerInp, &available);
	if ((err != noErr) || (available <= 0)) goto exit;

	oldPos = fPos;
	do {
		/* Get the highest number of bytes that will fit in our buffer */
		if ((available + fBufferCount) > BUFFER_LENGTH) {
			count = BUFFER_LENGTH - fBufferCount;
		}
		else {
			count = available;
		}
		err = FSRead(fSerInp, &count, fBuffer+fBufferCount);
		if (err != noErr) {
			/* Serial port error, this is serious, bail out */
			fBufferCount = 0;
			goto exit;
		}
		available -= count;
		fBufferCount += count;

		AccumulateBuffer(fBuffer, &fBufferCount);
	} while ((available > 0) && (err == noErr));

	/* Update the controller if the fPos values have changed */
	if ((oldPos.x != fPos.x) || (oldPos.y != fPos.y) || (oldPos.z != fPos.z)) {
		UpdateController();
	}
	
exit:
	return err; 
}
#endif

/*===========================================================================*\
 *
 *	Routine:	DynaTerm()
 *
 *	Comments:	Deactivates the controller, and ends communication with the
 *				device.
 *
\*===========================================================================*/

OSErr DynaTerm(
	void)
{
	OSErr err;

	/* Stop communication */
	err = DynaStopRunning();
	if (fConfiguration.serialPort != -1) {
		err = SerialPortTerm(fSerInp, fSerOut);
		fSerInp = 0;
		fSerOut = 0;
		fConfiguration.serialPort = -1;
	}

	/* Deactivate the controller */
	if (fControllerRef != NULL) {
		Q3Controller_Decommission(fControllerRef);
		fControllerRef = NULL;
	}

	/* Dispose serial ports name table */
	DisposeNameTable();

	return err;
}


/*===========================================================================*\
 *
 *	Routine:	DynaStartRunning()
 *
 *	Comments:	Starts the driver in the current operation mode
 *
\*===========================================================================*/

OSErr DynaStartRunning(
	void)
{
	OSErr err;
//	Byte command[1];

	if (fIsRunning) return noErr;
	if (fConfiguration.serialPort == -1) return err_NotResponding;
	
	fBufferCount = 0;
	err = SerRecvFlush(fSerInp);
	if (err != noErr) goto exit;

#if POLL_MODE
	if (fConfiguration.operationMode == oper_Stream) {
#endif
		/* Set up the deferred task block */
		fDefTask.qType		= dtQType;
		fDefTask.dtAddr		= NewDeferredTaskProc(DeferredDecode);
		fDefTask.dtParam	= 0;
		fDefTask.dtReserved	= 0;
		
		/* Set up the serial parameter block */
		fSerParm.ioCompletion	= NewIOCompletionProc(ASyncCompletion);
		fSerParm.ioRefNum		= fSerInp;
		fSerParm.ioPosMode		= fsAtMark;
		fSerParm.ioPosOffset	= 0;
		fSerParm.ioBuffer		= (Ptr) fBuffer;
		fSerParm.ioReqCount		= PACKET_LENGTH;
			
		/* Wait for the first package */
		err = PBReadAsync((ParmBlkPtr) &fSerParm);
		if (err != noErr) goto exit;
#if POLL_MODE
	}
#endif
	fIsRunning = true;

exit:
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	DynaStopRunning()
 *
 *	Comments:	Stops the driver
 *
\*===========================================================================*/

OSErr DynaStopRunning(
	void)
{
	OSErr err = noErr;

	if (!fIsRunning) goto exit;
#if POLL_MODE
	if (fConfiguration.operationMode == oper_Stream) {
#endif
		KillIO(fSerInp);
		KillIO(fSerOut);
		DisposeRoutineDescriptor(fDefTask.dtAddr);
		DisposeRoutineDescriptor(fSerParm.ioCompletion);
#if POLL_MODE
	}
#endif
	fIsRunning = false;

exit:
	return noErr;
}


/*===========================================================================*\
 *
 *	Routine:	DynaSetConfiguration()
 *
 *	Comments:	Set new driver configuration
 *				This will stop the driver if it was running
 *
\*===========================================================================*/

OSErr DynaSetConfiguration(
	TDynaConfiguration *config)
{
	OSErr err = noErr;

	/* Stop the driver */
	err = DynaStopRunning();
	if (err != noErr) goto exit;
	/* Set serial port */
	err = SerialPortSet(config->serialPort);
	if (err != noErr) goto exit;
	/* Set the rest of the configuration */
	fConfiguration = *config;
	
exit:
	return err;
}


/*===========================================================================*\
 *
 *	Routine:	DynaGetConfiguration()
 *
 *	Comments:	Get current driver configuration
 *
\*===========================================================================*/

void DynaGetConfiguration(
	TDynaConfiguration *config)
{
	*config = fConfiguration;
}


/*===========================================================================*\
 *
 *	Routine:	DynaGetIsRunning()
 *
 *	Comments:	Get current driver state
 *
\*===========================================================================*/

Boolean DynaGetIsRunning(
	void)
{
	return fIsRunning;
}




