/******************************************************************************
 **																			 **
 ** 	Module:		DriverShell.c											 **
 ** 																		 **
 ** 																		 **
 ** 	Purpose: 	Driver shell for Dynasight device						 **
 **					Runs as a Macintosh application, mainly so that it can	 **
 ** 				have a Quit menu item.									 **
 **					Created by Maarten Gribnau								 **
 **					email: m.w.gribnau@io.tudelft.nl						 **
 **					Loosely based on code for Magellan driver.				 **
 ** 				Copyright (C) 1992-96 Apple Computer, Inc.				 **
 **					All rights reserved.									 **
 ** 																		 **
 ** 																		 **
 *****************************************************************************/

#include <Controls.h>
#include <Devices.h>
#include <Dialogs.h>
#include <Fonts.h>
#include <Processes.h>
#include <Sound.h>
#include <Strings.h>
#include <ToolUtils.h>
#include <stdio.h>
#include <string.h>
#include <QD3D.h>
#include <QD3DErrors.h>

#include "DriverDynasight.h"
#include "PrefFile.h"
#include "SerialDriverNames.h"

// Global variable
TPreferences gPreferences;
FSSpec gPrefFile;
const OSType  Creator_Preferences	= 'MG21';

/******************************************************************************
 **																			 **
 **									Constants								 **
 **																			 **
 *****************************************************************************/

enum {									/* MENU resource IDs */
	menuApple = 128,
	menuFile,
	menuDriver = 256
};


enum {									/* Apple menu items */
	appleAbout = 1
};


enum {									/* File menu items */
	fileSettings = 1,
	fileQuit
};


enum {									/* Alert resource IDs */
	alrtQD3DError = 255,
	alrtQD3DWarning = 256,
	alrtQD3DNotice = 257,
	alrtOSError = 1001,
	alrtDriverError = 1002
};

enum {									/* Dialog resource IDs */
	dlgAbout = 1000,
	dlgSettings = 2000
};

enum {									/* Settings dialog control IDs */
	butOk = 1,
	butCancel = 2,
	cntlPort = 3,
	editOffsetX = 7,
	editOffsetY = 8,
	editOffsetZ = 9,
	editScaleX = 13,
	editScaleY = 14,
	editScaleZ = 15,
#if POLL_MODE
	radioPoll = 20,
	radioStream = 21,
	checkControlCursor = 22
#else
	checkControlCursor = 19
#endif
};



/******************************************************************************
 **																			 **
 **							Variables and Forward Declarations				 **
 **																			 **
 *****************************************************************************/

static MenuHandle fMenuApple;
static MenuHandle fMenuFile;
static EventRecord fEvent;
static Boolean fExit;
static TQ3Error			fFirstError		= kQ3ErrorNone;
static TQ3Error			fLastError		= kQ3ErrorNone;
static unsigned long	fNumErrors		= 0;
static TQ3Warning		fFirstWarning	= kQ3WarningNone;
static TQ3Warning		fLastWarning	= kQ3WarningNone;
static unsigned long	fNumWarnings	= 0;
static TQ3Notice		fFirstNotice	= kQ3NoticeNone;
static TQ3Notice		fLastNotice		= kQ3NoticeNone;
static unsigned long	fNumNotices		= 0;

static pascal Boolean OkCancelDlgFilter(
	DialogPtr theDlg,
	EventRecord *theEvent,
	short *itemHit);

static void FlipItem(
	DialogPtr dlg,
	short item);

/*static void SetItemByte(
	DialogPtr dlg,
	short item,
	Byte value);*/

static void SetItemFloat(
	DialogPtr dlg,
	short item,
	float value);

static void GetItemFloat(
	DialogPtr dlg,
	short item,
	float *value);

static void BuildDriverMenu(
	MenuHandle menu);

static void ClickInWindow(
	WindowPtr wind,
	Point where);

static void	ErrorAlert(
	OSErr err);

static void DoAppleAbout(
	void);

static void DoAppleDeskAcc(
	int item);

void DoFileSettings(
	void);

static void DoFileQuit(
	void);

static void DoMenuCommand(
	long itemVal);

static void WindowDrag(
	WindowPtr wind,
	Point where);

static void DoEventMouseDown(
	void);

static void DoEventKeyDown(
	void);

static void DoEventUpdate(
	WindowPtr wind);

static void DoEventActivate(
	WindowPtr wind);

static void Init(
	void);

static void Loop(	
	void);

static void Term(
	void);

void ShowError();
void ShowWarning();
void ShowNotice();
void ErrorMethod(
	TQ3Error firstError,
	TQ3Error lastError,
	long reference);
void WarningMethod(
	TQ3Warning firstWarning,
	TQ3Warning lastWarning,
	long reference);
void NoticeMethod(
	TQ3Notice firstNotice,
	TQ3Notice lastNotice,
	long reference);

void main(
	void);



/******************************************************************************
 **																			 **
 **								Basic Mac Glue Routines						 **
 **																			 **
 *****************************************************************************/


/*===========================================================================*\
 *
 *	Routine:	OkCancelDlgFilter()
 *
 *	Comments:	Dlg filter for default buttons Ok and Cancel
 *
\*===========================================================================*/

pascal Boolean OkCancelDlgFilter(
	DialogPtr /*theDlg*/,
	EventRecord *theEvent,
	short *itemHit)
{
	if (theEvent->what == keyDown) {
		switch ((theEvent->message) & charCodeMask) {
			case 0x0D:
			case 0x03:
				*itemHit = 1;
				return true;
			case 0x1B:
				*itemHit = 2;
				return true;
		}
	}
	return false;
}


/*===========================================================================*\
 *
 *	Routine:	FlipItem()
 *
 *	Comments:	Useful little subroutine for inverting dialog button values.
 *
\*===========================================================================*/

void FlipItem(
	DialogPtr dlg,
	short item)
{
	short iType;
	Rect iRect;
	ControlHandle iHandle;

	GetDialogItem(dlg, item, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, !GetControlValue(iHandle));
}


/*===========================================================================*\
 *
 *	Routine:	SetItemByte()
 *
 *	Comments:	Useful little subroutine for setting dialog text edit values.
 *
\*===========================================================================*/
/*
void SetItemByte(
	DialogPtr dlg,
	short item,
	Byte value)
{
	short iType;
	Rect iRect;
	ControlHandle iHandle;
	Str15 text;

	GetDialogItem(dlg, item, &iType, (Handle *)&iHandle, &iRect);
	if (iType & statText) {
		sprintf(((char *)text)+1, "%X", value);
		text[0] = strlen(((char *)text)+1);
		SetIText((Handle)iHandle, text);
	}
}
*/

/*===========================================================================*\
 *
 *	Routine:	SetItemFloat()
 *
 *	Comments:	Useful little subroutine for setting dialog text edit values.
 *
\*===========================================================================*/

void SetItemFloat(
	DialogPtr dlg,
	short item,
	float value)
{
	short iType;
	Rect iRect;
	ControlHandle iHandle;
	Str255 text;

	GetDialogItem(dlg, item, &iType, (Handle *)&iHandle, &iRect);
	if (iType & editText) {
		sprintf(((char *)text)+1, "%.6f", value);
		text[0] = strlen(((char *)text)+1);
		SetDialogItemText((Handle)iHandle, text);
	}
}


/*===========================================================================*\
 *
 *	Routine:	GetItemFloat()
 *
 *	Comments:	Useful little subroutine for setting dialog text edit values.
 *
\*===========================================================================*/

void GetItemFloat(
	DialogPtr dlg,
	short item,
	float *value)
{
	short iType;
	Rect iRect;
	ControlHandle iHandle;
	Str255 text;

	GetDialogItem(dlg, item, &iType, (Handle *)&iHandle, &iRect);
	if (iType & editText) {
		GetDialogItemText((Handle)iHandle, text);
		p2cstr(text);
		//text[text[0]] = '\0';
		sscanf((char *)text, "%f", value);
	}
}


/*===========================================================================*\
 *
 *	Routine:	BuildDriverMenu()
 *
 *	Comments:	Build a menu with the names of all serial drivers on the system
 *
\*===========================================================================*/

void
BuildDriverMenu(
	MenuHandle menu)
{
	short i, numDriverNames;
	Str255 driverName;
	
	if (GetNumDriverNames(&numDriverNames) != noErr) return;
	for (i = 0; i < numDriverNames; i++) {
		if (GetDriverName(i, driverName) == noErr) {
			AppendMenu(menu, driverName);
		}
		else {
			return;
		}
	}
}


/*===========================================================================*\
 *
 *	Routine:	ClickInWindow()
 *
 *	Comments:	Tracks a mouse click.
 *
\*===========================================================================*/

void ClickInWindow (
	WindowPtr  /* wind */,
	Point /* where */)
{
}


/*===========================================================================*\
 *
 *	Routine:	ErrorAlert()
 *
 *	Comments:	Puts up a error alert with an error code
 *
\*===========================================================================*/

void ErrorAlert(
	OSErr err)
{
	unsigned char numString[12];

	NumToString(err, numString);
	ParamText(numString, nil, nil, nil);
	if (err < 0) {
		/* System error (probably serial port) */
		StopAlert(alrtOSError, nil);
	}
	else {
		/* Owl driver error */
		StopAlert(alrtDriverError, nil);
	}
}


/*===========================================================================*\
 *
 *	Routine:	DoAppleAbout()
 *
 *	Comments:	Puts up the "About" dialog.
 *
\*===========================================================================*/

// ================= the event filter proc ==================
pascal Boolean AboutDlgFilter(
	DialogPtr theDlg,
	EventRecord *theEvent,
	short *itemHit);
pascal Boolean AboutDlgFilter(
	DialogPtr /*theDlg*/,
	EventRecord *theEvent,
	short *itemHit)
{
	if (theEvent->what == mouseDown) {
		*itemHit = 1;
		return true;
	}
	return false;
}
void DoAppleAbout (
	void)
{
	DialogPtr dlg;
	UniversalProcPtr modalFilterProc;
	short itemHit;
	Boolean done;

	dlg = GetNewDialog(dlgAbout, 0, (WindowPtr)-1);
	modalFilterProc = NewModalFilterProc(AboutDlgFilter);
	ShowWindow((WindowPtr) dlg);

	done = false;
	while (done == false) {
		ModalDialog(nil, /*modalFilterProc, */&itemHit);
		if (itemHit == 1) {
			done = true;
		}
	}
	DisposeDialog(dlg);
	DisposeRoutineDescriptor(modalFilterProc);
}


/*===========================================================================*\
 *
 *	Routine:	DoAppleDeskAcc()
 *
 *	Comments:	Invokes a desk accessory.
 *
\*===========================================================================*/

void DoAppleDeskAcc(
	int item)
{
	Str255 name;
	
	GetMenuItemText(fMenuApple, item, name);
	OpenDeskAcc(name);
}


/*===========================================================================*\
 *
 *	Routine:	DoFileSettings()
 *
 *	Comments:	Responds to a settings command.
 *
\*===========================================================================*/

void DoFileSettings(
	void)
{
	OSErr err;
	DialogPtr dlg;
	short	itemHit;
	Boolean done, doSet;
	short iType;
	Rect iRect;
	ControlHandle iHandle;
	MenuHandle menu;
	UniversalProcPtr modalFilterProc;
	TDynaConfiguration config;

	err = DynaStopRunning();
	if (err != noErr) goto exit;
	DynaGetConfiguration(&config);

	/* Create the driver pop up menu */
	menu = GetMenu(menuDriver);
	BuildDriverMenu(menu);

	/* Set up the dialog with the current values for the radio buttons */
	dlg = GetNewDialog(dlgSettings, 0, (WindowPtr)-1);
	modalFilterProc = NewModalFilterProc(OkCancelDlgFilter);
	GetDialogItem(dlg, cntlPort, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, config.serialPort+2);
	SetItemFloat(dlg, editOffsetX, config.offsetX);
	SetItemFloat(dlg, editOffsetY, config.offsetY);
	SetItemFloat(dlg, editOffsetZ, config.offsetZ);
	SetItemFloat(dlg, editScaleX, config.scaleX);
	SetItemFloat(dlg, editScaleY, config.scaleY);
	SetItemFloat(dlg, editScaleZ, config.scaleZ);
#if POLL_MODE
	GetDialogItem(dlg, radioPoll, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, config.operationMode == oper_Poll);
	GetDialogItem(dlg, radioStream, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, config.operationMode != oper_Poll);
#endif
	GetDialogItem(dlg, checkControlCursor, &iType, (Handle *)&iHandle, &iRect);
	SetControlValue(iHandle, config.controlCursor);
	ShowWindow((WindowPtr) dlg);

	/* Do modal dialog loop */
	ShowWindow(dlg);
	done = false;
	doSet = false;
	while (done == false) {
			ModalDialog(modalFilterProc, &itemHit);
			switch (itemHit) {
#if POLL_MODE
				case radioPoll:
				case radioStream:
					GetDialogItem(dlg, radioPoll, &iType, (Handle *)&iHandle, &iRect);
					SetControlValue(iHandle, 0);
					GetDialogItem(dlg, radioStream, &iType, (Handle *)&iHandle, &iRect);
					SetControlValue(iHandle, 0);
					GetDialogItem(dlg, itemHit, &iType, (Handle *)&iHandle, &iRect);
					SetControlValue(iHandle, 1);
					break;
#endif
				case checkControlCursor:
					FlipItem(dlg, itemHit);
					break;
				
				case butOk:
					/* Get the current settings and activate them */
					GetDialogItem(dlg, cntlPort, &iType, (Handle *)&iHandle, &iRect);
					config.serialPort = GetControlValue(iHandle) - 2;
#if POLL_MODE
					GetDialogItem(dlg, radioPoll, &iType, (Handle *)&iHandle, &iRect);
					config.operationMode = GetControlValue(iHandle) ? oper_Poll : oper_Stream;
#endif
					GetDialogItem(dlg, checkControlCursor, &iType, (Handle *)&iHandle, &iRect);
					config.controlCursor = GetControlValue(iHandle);
					GetItemFloat(dlg, editOffsetX, &config.offsetX);
					GetItemFloat(dlg, editOffsetY, &config.offsetY);
					GetItemFloat(dlg, editOffsetZ, &config.offsetZ);
					GetItemFloat(dlg, editScaleX, &config.scaleX);
					GetItemFloat(dlg, editScaleY, &config.scaleY);
					GetItemFloat(dlg, editScaleZ, &config.scaleZ);
					done = true;
					doSet = true;
					break;
					
				case butCancel:
					done = true;
					break;
			}
	}
	
	/* Clean up */
	DisposeDialog(dlg);
	DisposeRoutineDescriptor(modalFilterProc);

	if (doSet) {
		err = DynaSetConfiguration(&config);
		if (err != noErr) goto exit;
	}
	if (config.serialPort != -1) {
		err = DynaStartRunning();
	}
	if (err != noErr) goto exit;
	
exit:
	if (err != noErr) {
		ErrorAlert(err);
	}
}


/*===========================================================================*\
 *
 *	Routine:	DoFileQuit()
 *
 *	Comments:	Responds to a Quit command.
 *
\*===========================================================================*/

void DoFileQuit(
	void)
{
	fExit = true;
}


/*===========================================================================*\
 *
 *	Routine:	DoMenuCommand()
 *
 *	Comments:	Switches out to handle a menu selection.
 *
\*===========================================================================*/

void DoMenuCommand(
	long itemVal)
{
	int item;
	
	item = LoWord(itemVal);
	
	switch (HiWord(itemVal)) {
		case menuApple:
			switch (item) {
				case appleAbout:
					DoAppleAbout();
					break;
				default:
					DoAppleDeskAcc(item);
					break;
			}
			break;
		
		case menuFile:
			switch (item) {
				case fileSettings:
					DoFileSettings();
					break;
				case fileQuit:
					DoFileQuit();
					break;
			}
			break;
	}
	
	HiliteMenu(0);
}


/*===========================================================================*\
 *
 *	Routine:	WindowDrag()
 *
 *	Comments:	Drags the window.
 *
\*===========================================================================*/

void WindowDrag(
	WindowPtr wind,
	Point where)
{
	DragWindow(wind, where, &qd.screenBits.bounds);
}


/*===========================================================================*\
 *
 *	Routine:	DoEventMouseDown()
 *
 *	Comments:	Responds to a mouse-down event in menu bar or system window.
 *
\*===========================================================================*/

void DoEventMouseDown(
	void)
{
	WindowPtr wind;
	
	switch (FindWindow(fEvent.where, &wind)) {
		case inMenuBar:
			DoMenuCommand(MenuSelect(fEvent.where));
			break;
		case inSysWindow:
			SystemClick(&fEvent, wind);
			break;
		case inContent:
			ClickInWindow(wind, fEvent.where);
			break;
		case inDrag:
			WindowDrag(wind, fEvent.where);
			break;
	}
}


/*===========================================================================*\
 *
 *	Routine:	DoEventKeyDown()
 *
 *	Comments:	Processes a key down by checking for a menu key.
 *
\*===========================================================================*/

void DoEventKeyDown(
	void)
{
	char ch;
	
	ch = fEvent.message & charCodeMask;
	
	if (fEvent.modifiers & cmdKey)
		DoMenuCommand(MenuKey(ch));
}


/*===========================================================================*\
 *
 *	Routine:	DoEventUpdate()
 *
 *	Comments:	Processes an update event.
 *
\*===========================================================================*/

void DoEventUpdate(
	WindowPtr wind)
{
	BeginUpdate(wind);
	EndUpdate(wind);
}


/*===========================================================================*\
 *
 *	Routine:	DoEventActivate()
 *
 *	Comments:	Processes an activate event.
 *
\*===========================================================================*/

void DoEventActivate(
	WindowPtr wind)
{
	wind; /* unused */
}


/*===========================================================================*\
 *
 *	Routine:	Init()
 *
 *	Comments:	Sets up menus, etc.
 *
\*===========================================================================*/

void Init(
	void)
{
	OSErr err;
	TQ3Status status;
	
	fExit = false;
	MaxApplZone();
	
	InitGraf(&qd.thePort);
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(nil);
	InitCursor();
	
	GetNextEvent(everyEvent, &fEvent);
	GetNextEvent(everyEvent, &fEvent);
	GetNextEvent(everyEvent, &fEvent);
	
	/* Form the menubar */
	InsertMenu(fMenuApple = GetMenu(menuApple), 0);
	InsertMenu(fMenuFile = GetMenu(menuFile), 0);
	AppendResMenu(fMenuApple, 'DRVR');
	DrawMenuBar();
	
	SetCursor(&qd.arrow);
	
	/* Create the serial driver name table */
	NewNameTable();

	/* Load the preferences of the driver */
	GetPreferencesFile(Creator_Preferences, "\pDynasight Preferences", &gPrefFile);
	LoadPreferences(&gPrefFile, &gPreferences);
	
	/* Initialize Escher */
	Q3Initialize();
	
	/* Set the error, warning and notice methods */
	status = Q3Error_Register(ErrorMethod, 0);
	if (status != kQ3Success) { err = err_QD3DInitialise; goto exit; }
	status = Q3Warning_Register(WarningMethod, 0);
	if (status != kQ3Success) { err = err_QD3DInitialise; goto exit; }
	status = Q3Notice_Register(NoticeMethod, 0);
	if (status != kQ3Success) { err = err_QD3DInitialise; goto exit; }
	
	/* Initialize the device */
	err = DynaInit(&gPreferences);
	if (err != noErr) { goto exit; }
	err = DynaStartRunning();
	if (err != noErr) { goto exit; }

exit:
	if (err != noErr) {
		ErrorAlert(err);
	}
}


/*===========================================================================*\
 *
 *	Routine:	Loop()
 *
 *	Comments:	Gets and processes events.
 *
\*===========================================================================*/

void Loop(	
	void)
{
#if POLL_MODE
	OSErr err;
	TDynaConfiguration config;
#endif
	long sleep;
	
	sleep = 0;
	while (!fExit) {
		WaitNextEvent(everyEvent, &fEvent, sleep, nil);
		
		switch(fEvent.what) {
			case mouseDown:
				DoEventMouseDown();
				break;
			case keyDown:
				DoEventKeyDown();
				break;
			case activateEvt:
				DoEventActivate((WindowPtr) fEvent.message);
				break;
			case updateEvt:
				DoEventUpdate((WindowPtr) fEvent.message);
				break;
		}
#if POLL_MODE
		DynaGetConfiguration(&config);
		if (config.operationMode == oper_Poll) {
			err = DynaPoll();
			sleep = 1;
		}
		else {
			sleep = 60;
		}
#else
		sleep = 60;
#endif
		if (fLastError != kQ3WarningNone) {
			ShowError();
			fFirstError = kQ3ErrorNone;
			fLastError = kQ3ErrorNone;
		}
		if (fLastWarning != kQ3WarningNone) {
			ShowWarning();
			fFirstWarning = kQ3WarningNone;
			fLastWarning = kQ3WarningNone;
		}
		if (fLastNotice != kQ3NoticeNone) {
			ShowNotice();
			fFirstNotice = kQ3NoticeNone;
			fLastNotice	= kQ3NoticeNone;
		}
	}
}


/*===========================================================================*\
 *
 *	Routine:	Term()
 *
 *	Comments:	Cleans up before exit.
 *
\*===========================================================================*/

void Term(
	void)
{
	DynaGetConfiguration(&gPreferences);
	SavePreferences(&gPrefFile, &gPreferences);
	DynaTerm();
	
	Q3Exit();

	/* Dispose the serial driver name table */
	DisposeNameTable();
}


void ErrorMethod(
	TQ3Error firstError,
	TQ3Error lastError,
	long /*reference*/)
{
	fFirstError = firstError;
	fLastError = lastError;
	fNumErrors++;
}
void WarningMethod(
	TQ3Warning firstWarning,
	TQ3Warning lastWarning,
	long /*reference*/)
{
	fFirstWarning = firstWarning;
	fLastWarning = lastWarning;
	fNumWarnings++;
}
void NoticeMethod(
	TQ3Notice firstNotice,
	TQ3Notice lastNotice,
	long /*reference*/)
{
	fFirstNotice = firstNotice;
	fLastNotice = lastNotice;
	fNumNotices++;
}


void ShowError()
{
	short answer;
	Str15 first, last, num;
	
	NumToString(fFirstError, first);
	NumToString(fLastError, last);
	NumToString(fNumErrors, num);
	ParamText(first, last, num, NULL);
	answer = StopAlert(alrtQD3DError, NULL);
	if (answer == 2) {
		DynaTerm();
		Q3Exit();
		ExitToShell();
	}
}
void ShowWarning()
{
	short answer;
	Str15 first, last, num;
	
	NumToString(fFirstWarning, first);
	NumToString(fLastWarning, last);
	NumToString(fNumWarnings, num);
	ParamText(first, last, num, NULL);
	answer = CautionAlert(alrtQD3DWarning, NULL);
	if (answer == 2) {
		DynaTerm();
		Q3Exit();
		ExitToShell();
	}
}
void ShowNotice()
{
	short answer;
	Str15 first, last, num;
	
	NumToString(fFirstNotice, first);
	NumToString(fLastNotice, last);
	NumToString(fNumNotices, num);
	ParamText(first, last, num, NULL);
	answer = NoteAlert(alrtQD3DNotice, NULL);
	if (answer == 2) {
		DynaTerm();
		Q3Exit();
		ExitToShell();
	}
}

/*===========================================================================*\
 *
 *	Routine:	main()
 *
 *	Comments:	xxx
 *
\*===========================================================================*/

void main(
	void)
{
	printf("moof!");
	Init();
	printf("moof!");
//	ErrorMethod(kQ3ErrorNone, kQ3ErrorNotInitialized, 0);
//	WarningMethod(kQ3WarningNone, kQ3WarningNoObjectSupportForDrawMethod, 0);
//	NoticeMethod(kQ3NoticeNone, kQ3NoticeDataAlreadyEmpty, 0);
	Loop();
	Term();
}